#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class dmol3_section_hirshfeld_population(MSection):
    '''
    Hirshfeld Population Analysis Section
    '''

    m_def = Section(validate=False)

    dmol3_hirshfeld_population = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Hirshfeld Population Analysis
        ''')


class dmol3_section_mulliken_population(MSection):
    '''
    Mulliken Population Analysis Section
    '''

    m_def = Section(validate=False)

    dmol3_mulliken_population = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Mulliken Population Analysis
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    dmol3_aux_density = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 aux density
        ''')

    dmol3_aux_partition = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 aux parition
        ''')

    dmol3_basis_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 basis name
        ''')

    dmol3_calculation_type = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 calculation type
        ''')

    dmol3_charge = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 system charge
        ''')

    dmol3_electrostatic_moments = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Electrostatic_Moments
        ''')

    dmol3_functional_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 functional name
        ''')

    dmol3_hirshfeld_analysis = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Hirshfeld_Analysis
        ''')

    dmol3_integration_grid = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 integration grid
        ''')

    dmol3_kpoints = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Kpoints
        ''')

    dmol3_mulliken_analysis = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Mulliken_Analysis
        ''')

    dmol3_nuclear_efg = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Nuclear_EFG
        ''')

    dmol3_occupation_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Occupation name
        ''')

    dmol3_occupation_width = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 Occupation width
        ''')

    dmol3_opt_coordinate_system = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 OPT_Coordinate_System
        ''')

    dmol3_opt_displacement_convergence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Displacement_Convergence
        ''')

    dmol3_opt_energy_convergence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Energy_Convergence
        ''')

    dmol3_opt_gdiis = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 OPT_Gdiis
        ''')

    dmol3_opt_gradient_convergence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Gradient_Convergence
        ''')

    dmol3_opt_hessian_project = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 OPT_Hessian_Project
        ''')

    dmol3_opt_iterations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 OPT_Iterations
        ''')

    dmol3_opt_max_displacement = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Max_Displacement
        ''')

    dmol3_opt_steep_tol = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 OPT_Steep_Tol
        ''')

    dmol3_optical_absorption = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Optical_Absorption
        ''')

    dmol3_partial_dos = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 Partial_Dos
        ''')

    dmol3_pseudopotential_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 pseudopotential name
        ''')

    dmol3_rcut = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 atom R_cut
        ''')

    dmol3_scf_charge_mixing = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 SCF_Charge_Mixing
        ''')

    dmol3_scf_density_convergence = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 SCF_Density_Convergence
        ''')

    dmol3_scf_diis_name = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 SCF_DIIS name
        ''')

    dmol3_scf_diis_number = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 SCF_DIIS number
        ''')

    dmol3_scf_direct = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 SCF_Direct
        ''')

    dmol3_scf_iterations = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 SCF_Iterations
        ''')

    dmol3_scf_number_bad_steps = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 SCF_Number_Bad_Steps
        ''')

    dmol3_scf_restart = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 SCF_Restart
        ''')

    dmol3_scf_spin_mixing = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 SCF_Spin_Mixing
        ''')

    dmol3_spin_polarization = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 spin polarization
        ''')

    dmol3_spin = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 number of unpaired electrons
        ''')

    dmol3_symmetry = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 sysmmetry
        ''')


class ScfIteration(simulation.calculation.ScfIteration):

    m_def = Section(validate=False, extends_base_section=True)

    dmol3_binding_energy_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        dmol3 binding energy at every SCF
        ''')

    dmol3_convergence_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 convergence at every SCF
        ''')

    dmol3_number_scf_iteration = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        dmol3 iteration number at every SCF
        ''')

    dmol3_time_scf_iteration = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        dmol3 time at every SCF
        ''')


class BandEnergies(simulation.calculation.BandEnergies):

    m_def = Section(validate=False, extends_base_section=True)

    dmol3_eigenvalue_eigenvalue = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='joule',
        description='''
        Single eigenvalue
        ''')

    dmol3_eigenvalue_occupation = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Occupation of single eigenfunction
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    dmol3_geometry_atom_labels = Quantity(
        type=str,
        shape=[],
        description='''
        labels of atom
        ''')

    dmol3_geometry_atom_positions_x = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        x component of atomic position
        ''')

    dmol3_geometry_atom_positions_y = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        y component of atomic position
        ''')

    dmol3_geometry_atom_positions_z = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        z component of atomic position
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    dmol3_program_compilation_date = Quantity(
        type=str,
        shape=[],
        description='''
        dmol3 compilation date
        ''')

    dmol3_program_compilation_time = Quantity(
        type=str,
        shape=[],
        description='''
        dmol compilation date
        ''')


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    dmol3_section_hirshfeld_population = SubSection(
        sub_section=SectionProxy('dmol3_section_hirshfeld_population'),
        repeats=True)

    dmol3_section_mulliken_population = SubSection(
        sub_section=SectionProxy('dmol3_section_mulliken_population'),
        repeats=True)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
from .amber.parser import AmberParser
from .asap.parser import AsapParser
from .dftbplus.parser import DFTBPlusParser
from .dlpoly.parser import DLPolyParser
from .gromacs.parser import GromacsParser
from .gromos.parser import GromosParser
from .gulp.parser import GulpParser
from .lammps.parser import LammpsParser
from .libatoms.parser import LibAtomsParser
from .mopac.parser import MopacParser
from .namd.parser import NAMDParser
from .openkim.parser import OpenKIMParser
from .tinker.parser import TinkerParser



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation
from nomad.datamodel.metainfo import workflow


m_package = Package()


class x_amber_mdin_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category()


class x_amber_mdout_single_configuration_calculation(MCategory):
    '''
    Parameters of mdout belonging to section_single_configuration_calculation.
    '''

    m_def = Category()


class x_amber_mdout_method(MCategory):
    '''
    Parameters of mdin belonging to section method.
    '''

    m_def = Category()


class x_amber_mdout_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category()


class x_amber_mdin_run(MCategory):
    '''
    Parameters of mdin belonging to settings run.
    '''

    m_def = Category()


class x_amber_section_input_output_files(MSection):
    '''
    Temperory variable to store input and output file keywords
    '''

    m_def = Section(validate=False)


class x_amber_section_single_configuration_calculation(MSection):
    '''
    section for gathering values for MD steps
    '''

    m_def = Section(validate=False)


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_amber_atom_positions_image_index = Quantity(
        type=np.dtype(np.int32),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        PBC image flag index.
        ''')

    x_amber_atom_positions_scaled = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='dimensionless',
        description='''
        Position of the atoms in a scaled format [0, 1].
        ''')

    x_amber_atom_positions_wrapped = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_atoms', 3],
        unit='meter',
        description='''
        Position of the atoms wrapped back to the periodic box.
        ''')

    x_amber_dummy = Quantity(
        type=str,
        shape=[],
        description='''
        dummy
        ''')

    x_amber_mdin_finline = Quantity(
        type=str,
        shape=[],
        description='''
        finline in mdin
        ''')

    x_amber_traj_timestep_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_traj_number_of_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_traj_box_bound_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_traj_box_bounds_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_traj_variables_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_traj_atoms_store = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')


class MolecularDynamics(workflow.MolecularDynamics):

    m_def = Section(validate=False, extends_base_section=True)

    x_amber_barostat_target_pressure = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='pascal',
        description='''
        MD barostat target pressure.
        ''')

    x_amber_barostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD barostat relaxation time.
        ''')

    x_amber_barostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD barostat type, valid values are defined in the barostat_type wiki page.
        ''')

    x_amber_integrator_dt = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD integration time step.
        ''')

    x_amber_integrator_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD integrator type, valid values are defined in the integrator_type wiki page.
        ''')

    x_amber_periodicity_type = Quantity(
        type=str,
        shape=[],
        description='''
        Periodic boundary condition type in the sampling (non-PBC or PBC).
        ''')

    x_amber_langevin_gamma = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        Langevin thermostat damping factor.
        ''')

    x_amber_number_of_steps_requested = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        Number of requested MD integration time steps.
        ''')

    x_amber_thermostat_level = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat level (see wiki: single, multiple, regional).
        ''')

    x_amber_thermostat_target_temperature = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='kelvin',
        description='''
        MD thermostat target temperature.
        ''')

    x_amber_thermostat_tau = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='second',
        description='''
        MD thermostat relaxation time.
        ''')

    x_amber_thermostat_type = Quantity(
        type=str,
        shape=[],
        description='''
        MD thermostat type, valid values are defined in the thermostat_type wiki page.
        ''')


class AtomParameters(simulation.method.AtomParameters):

    m_def = Section(validate=False, extends_base_section=True)

    x_amber_atom_type_element = Quantity(
        type=str,
        shape=[],
        description='''
        Element symbol of an atom type.
        ''')

    x_amber_atom_type_radius = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        van der Waals radius of an atom type.
        ''')


class Interaction(simulation.method.Interaction):

    m_def = Section(validate=False, extends_base_section=True)

    x_amber_interaction_atom_to_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each interaction atoms.
        ''')

    x_amber_number_of_defined_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions (L-J pairs).
        ''')

    x_amber_pair_interaction_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['x_amber_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions.
        ''')

    x_amber_pair_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['x_amber_number_of_defined_pair_interactions', 2],
        description='''
        Pair interactions parameters.
        ''')

    x_amber_molecule_interaction_atom_to_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type of each molecule interaction atoms.
        ''')

    x_amber_number_of_defined_molecule_pair_interactions = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        Number of defined pair interactions within a molecule (L-J pairs).
        ''')

    x_amber_pair_molecule_interaction_parameters = Quantity(
        type=np.dtype(np.float64),
        shape=['number_of_defined_molecule_pair_interactions', 2],
        description='''
        Molecule pair interactions parameters.
        ''')

    x_amber_pair_molecule_interaction_to_atom_type_ref = Quantity(
        type=simulation.method.AtomParameters,
        shape=['x_amber_number_of_defined_pair_interactions', 'number_of_atoms_per_interaction'],
        description='''
        Reference to the atom type for pair interactions within a molecule.
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_amber_program_version_date = Quantity(
        type=str,
        shape=[],
        description='''
        Program version date.
        ''')

    x_amber_xlo_xhi = Quantity(
        type=str,
        shape=[],
        description='''
        test
        ''')

    x_amber_data_file_store = Quantity(
        type=str,
        shape=[],
        description='''
        Filename of data file
        ''')

    x_amber_program_working_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_program_execution_host = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_program_execution_path = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_program_module = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_program_execution_date = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_program_execution_time = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_mdin_header = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_mdin_wt = Quantity(
        type=str,
        shape=[],
        description='''
        tmp
        ''')

    x_amber_section_input_output_files = SubSection(
        sub_section=SectionProxy('x_amber_section_input_output_files'),
        repeats=False)


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_amber_section_single_configuration_calculation = SubSection(
        sub_section=SectionProxy('x_amber_section_single_configuration_calculation'),
        repeats=True)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import numpy as np
import logging
from datetime import datetime
from netCDF4 import Dataset  # pylint: disable=no-name-in-module
from ase.data import chemical_symbols

from nomad.parsing.file_parser import TextParser, Quantity, FileParser
from nomad.units import ureg
from nomad.datamodel.metainfo.simulation.run import Run, Program, TimeRun
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from nomad.datamodel.metainfo.simulation.method import Method
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Energy, EnergyEntry, BandEnergies)
from .metainfo.yambo import (
    x_yambo_dipoles, x_yambo_dynamic_dielectric_matrix_fragment, x_yambo_io,
    x_yambo_dynamic_dielectric_matrix, x_yambo_local_xc_nonlocal_fock,
    x_yambo_dyson, x_yambo_local_xc_nonlocal_fock_bandenergies, x_yambo_bare_xc,
    x_yambo_bare_xc_bandenergies, x_yambo_module, x_yambo_transferred_momenta
)


class MainfileParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        re_f = r'[-+]*\d*\.\d+[Ee]*[-+]*\d*'

        io_quantities = [
            Quantity(
                'key_value',
                r'([A-Z\d].+?)(?:\(.+\)|\[.+\]| |)(:.+?)(?:\[|\n)',
                str_operation=lambda x: [v.strip() for v in x.split(':')],
                repeats=True
            ),
            Quantity('file', r'\[(?:RD|WR)(.+?)\]', dtype=str),
            Quantity('sn', r'- S/N *(\d+)', dtype=str)
        ]

        energies_quantities = [
            Quantity(
                'fermi',
                rf'Fermi Level.+?: +({re_f})',
                dtype=np.float64, unit=ureg.eV
            ),
            Quantity(
                'conduction',
                rf'Conduction Band Min +: +({re_f})',
                dtpye=np.float64, unit=ureg.eV
            ),
            Quantity(
                'valence',
                rf'Valence Band Max +: +({re_f})',
                dtpye=np.float64, unit=ureg.eV
            ),
            Quantity(
                'valence_conduction',
                rf'VBM / CBm +\[ev\]: +({re_f}) +({re_f})',
                dtpye=np.dtype(np.float64), unit=ureg.eV
            ),
            Quantity(
                'x_yambo_filled_bands',
                r'Filled Bands +: +(\d+)', dtype=np.int32, str_operation=lambda x: [1, int(x)]
            ),
            Quantity(
                'x_yambo_empty_bands',
                r'Empty Bands +: +([\d ]+)', dtype=np.dtype(np.int32)
            ),
            Quantity(
                'x_yambo_electronic_temperature',
                rf'Electronic Temp.+?: +{re_f} +({re_f})',
                dtype=np.float64, unit=ureg.kelvin
            ),
            Quantity(
                'x_yambo_bosonic_temperature',
                rf'Bosonic +Temp.+?: +{re_f} +({re_f})',
                dtype=np.float64, unit=ureg.kelvin
            ),
            Quantity(
                'x_yambo_finite_temperature_mode',
                r'Finite Temperature mode: +(\S+)', str_operation=lambda x: x == 'yes'
            ),
            Quantity(
                'x_yambo_electronic_density',
                rf'El\. density.+?: +(.+?)(?:\[|\n)',
                str_operation=lambda x: x.strip().split()[-1], dtype=np.float64
            ),
            Quantity(
                'states_summary',
                r'States summary +: Full +Metallic +Empty\s+(.+)',
                str_operation=lambda x: [v.split('-') for v in x.strip().split()]
            ),
            Quantity(
                'x_yambo_indirect_gaps',
                rf'Indirect Gaps.+?: +({re_f}) +({re_f})',
                dtype=np.dtype(np.float64), unit=ureg.eV
            ),
            Quantity(
                'x_yambo_direct_gaps',
                rf'Direct Gaps.+?: +({re_f}) +({re_f})',
                dtype=np.dtype(np.float64), unit=ureg.eV
            ),
            Quantity(
                'x_yambo_indirect_gap',
                rf'Indirect Gap.+?: +({re_f})',
                dtype=np.float64, unit=ureg.eV
            ),
            Quantity(
                'x_yambo_direct_gap',
                rf'Direct Gap.+?: +({re_f})',
                dtype=np.float64, unit=ureg.eV
            ),
            Quantity(
                'x_yambo_direct_gap_kpoint',
                r'Direct Gap localized at k-point.+?: +(\d+)',
                dtype=np.int32
            ),
            Quantity(
                'x_yambo_indirect_gap_kpoints',
                r'Indirect Gap between k-points.+?: +(\d+) +(\d+)',
                dtype=np.int32
            ),
        ]

        qp_properties_quantity = Quantity(
            'qp_properties',
            r'QP properties and I/O([\s\S]+? S/N \d+.+)',
            sub_parser=TextParser(quantities=[
                Quantity(
                    'qp_energy',
                    r'(QP \[eV\] @ K[\s\S]+?)\n *\n',
                    repeats=True, sub_parser=TextParser(quantities=[
                        Quantity(
                            'band',
                            rf'B= *(\d+) Eo= *({re_f}) E= *({re_f}) E-Eo= *({re_f}) '
                            rf'Re\(Z\)= *({re_f}) Im\(Z\)= *({re_f}) nlXC= *({re_f}) lXC= *({re_f}) So= *({re_f})',
                            repeats=True, dtype=np.dtype(np.float64)
                        ),
                        Quantity(
                            'kpoint',
                            r'K *\[\d+\].+?\: *(.+)',
                            dtype=np.dtype(np.float64)
                        )
                    ])
                ),
                Quantity(
                    'output',
                    r'(\[WR.+?\.QP\][\s\S]+?- S/N \d+.+)',
                    repeats=True, sub_parser=TextParser(quantities=io_quantities)
                )
            ])
        )

        module_quantities = [
            Quantity(
                'dipoles',
                r'Dipoles *\n([\s\S]+?)\n *\[\d+\]',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'input',
                        r'(\[RD.+?\][\s\S]+?- S/N \d+.+)',
                        repeats=True, sub_parser=TextParser(quantities=io_quantities)
                    ),
                    Quantity(
                        'output',
                        r'(\[WR.+?\.dipoles\][\s\S]+?- S/N \d+.+)',
                        repeats=True, sub_parser=TextParser(quantities=io_quantities)
                    )
                ])

            ),
            Quantity(
                'local_xc_nonlocal_fock',
                r'Local Exchange-Correlation \+ Non-Local Fock([\s\S]+?(?:\n *\[\d+\]|\Z))',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'output',
                        r'(\[WR.+?\.HF_and_locXC\][\s\S]+?- S/N \d+.+)',
                        repeats=True, sub_parser=TextParser(quantities=io_quantities)
                    ),
                    Quantity(
                        'x_yambo_plane_waves_vxc',
                        r'\[VXC\] Plane waves : *(\d+)',
                        dtype=np.int32
                    ),
                    Quantity(
                        'x_yambo_plane_waves_exs',
                        r'\[EXS\] Plane waves : *(\d+)',
                        dtype=np.int32
                    ),
                    Quantity(
                        'x_yambo_mesh_size',
                        r'Mesh size: *(\d+) *(\d+) *(\d+)',
                        dtype=np.dtype(np.int32)
                    ),
                    Quantity(
                        'energy_xc',
                        rf'E_xc *: *({re_f}) \[Ha\]',
                        dtype=np.float64, unit='hartree'
                    ),
                    Quantity(
                        'corrections',
                        r'Corrections @ K \[\d+\] *: *\[eV\]([\s\S]+?)\n *\n',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity(
                                'band',
                                rf'\<\d+\|nlXC\|\d+\> *= *({re_f}) *{re_f} \<\d+\|lXC\|\d+\> *= *({re_f}) *{re_f}',
                                repeats=True, dtype=np.dtype(np.float64)
                            )
                        ])
                    ),
                    Quantity(
                        'hf_occupations',
                        r'Hartree-Fock occupations report([\s\S]+?)(?:\n *\[\d+|\Z)',
                        sub_parser=TextParser(quantities=energies_quantities)
                    )
                ])
            ),
            # TODO add support for em1d
            Quantity(
                'dynamic_dielectric_matrix',
                r'Dynamic.+?Dielectric Matrix([\s\S]+?(?:\n *\[\d+\]|\Z))',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'output',
                        r'(\[WR.+?(?:pp|em1d|dip_iR_and_P)\][\s\S]+?- S/N \d+.+)',
                        repeats=True, sub_parser=TextParser(quantities=io_quantities)
                    ),
                    Quantity(
                        'x_yambo_mesh_size',
                        r'Mesh size: *(\d+) *(\d+) *(\d+)',
                        dtype=np.dtype(np.int32)
                    ),
                ])
            ),
            Quantity(
                'bare_xc',
                r'Bare local and non-local Exchange-Correlation([\s\S]+?(?:\n *\[\d+\]|\Z))',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'output',
                        r'(\[WR.+?\.HF_and_locXC\][\s\S]+?- S/N \d+.+)',
                        repeats=True, sub_parser=TextParser(quantities=io_quantities)
                    ),
                    Quantity(
                        'xc_hf_dft',
                        r'XC HF and DFT \[eV\]([\s\S]+?)\n *\n',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity(
                                'band',
                                rf'\<\d+\|HF\|\d+\> *= *({re_f}) *{re_f} *\<\d+\|DFT\|\d+\> *= *({re_f}) *{re_f}',
                                repeats=True, dtype=np.dtype(np.float64)
                            )
                        ])
                    ),
                    Quantity(
                        'hf_occupations',
                        r'HF occupations report([\s\S]+?Direct Gaps.+)',
                        sub_parser=TextParser(quantities=energies_quantities)
                    )
                ])
            ),
            Quantity(
                'dyson',
                r'Dyson equation: Newton solver([\s\S]+?(?:\n *\[\d+\]|\Z))',
                sub_parser=TextParser(quantities=[
                    qp_properties_quantity,
                    Quantity(
                        'g0w0',
                        r'G0W0([\s\S]+?\n *\[\d+\.\d+\])',
                        sub_parser=TextParser(quantities=[
                            Quantity(
                                'x_yambo_bands_range',
                                r'Bands range *: *(\d+) *(\d+)',
                                dtype=np.dtype(np.int32)
                            ),
                            Quantity(
                                'x_yambo_g_damping',
                                rf'G damping.+?: *({re_f})',
                                dtype=np.float64
                            ),
                            Quantity(
                                'x_yambo_mesh_size',
                                r'Mesh size: *(\d+) *(\d+) *(\d+)',
                                dtype=np.dtype(np.int32)
                            ),
                            Quantity(
                                'input',
                                r'(\[RD.+?\.pp\][\s\S]+?- S/N \d+.+)',
                                repeats=True, sub_parser=TextParser(quantities=io_quantities)
                            )
                        ])
                    ),
                ])
            )
        ]

        self._quantities = [
            Quantity(
                'version',
                r'Version ([\d.]+ Revision \d+)', flatten=False, dtype=str
            ),
            Quantity('hash', r'Hash (\S+)', dtype=str),
            Quantity('build', r'(\S+) Build', dtype=str),
            Quantity(
                'date_start',
                r' (\d\d/\d\d/\d\d\d\d) at (\d\d:\d\d) YAMBO @ .+',
                flatten=False, dtype=str
            ),
            Quantity(
                'cpu_files_io',
                r'((?:Cores |CPU structure)[\s\S]+?)\n *\[\d+\]',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'parameters',
                        r'([A-Z][\w/ ]+).*?(?: is | in |: ) *(\S+)',
                        repeats=True, str_operation=lambda x:[v.strip() for v in x.rsplit(' ', 1)]
                    ),
                    Quantity(
                        'input',
                        r'( \[RD.+[\s\S]+?- S/N \d+.+)',
                        repeats=False, sub_parser=TextParser(quantities=io_quantities)
                    )
                ])
            ),
            Quantity(
                'core_variables_setup',
                r'(CORE Variables Setup[\s\S]+?)\n *\[\d+\]',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'energies_occupations',
                        r'Energies.+?& Occupations([\s\S]+?)(?:\[0|\Z)',
                        sub_parser=TextParser(quantities=energies_quantities + [
                            Quantity(
                                'eigenenergies',
                                rf'Energy unit is electronVolt \[eV\]([\s\S]+E *{re_f} *{re_f} *{re_f}.+)',
                                sub_parser=TextParser(quantities=[
                                    Quantity(
                                        'energies',
                                        rf'\n *E *({re_f} .+)',
                                        repeats=True, dtype=np.dtype(np.float64)),
                                    Quantity(
                                        'kpoints',
                                        rf'({re_f} *{re_f} *{re_f}) \(rlu\)',
                                        repeats=True, dtype=np.dtype(np.float64)),
                                    Quantity(
                                        'kpoints_weights',
                                        rf'weight +({re_f})',
                                        repeats=True, dtype=np.float64)]))
                        ])
                    )
                ])
            ),
            Quantity(
                'transferred_momenta',
                r'Transferred momenta grid([\s\S]+?)\n *\[\d+\]',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'input',
                        r'( \[RD.+[\s\S]+?- S/N \d+.+)',
                        repeats=True, sub_parser=TextParser(quantities=io_quantities)
                    ),
                    Quantity(
                        'qpoints',
                        rf'Q \[\d+\] *: *({re_f}) *({re_f}) *({re_f}) *\(iku\) \* weight *({re_f})',
                        repeats=True, dtype=np.dtype(np.float64)
                    ),
                    Quantity(
                        'module',
                        r'((?:Dipoles *\n|Dynamic Dielectric|Dyson|Bare local|Local Exchange)[\s\S]+?\n *\[\d+\.\d+\])',
                        repeats=True, sub_parser=TextParser(quantities=module_quantities)
                    ),
                    qp_properties_quantity
                ])
            ),
            Quantity(
                'module',
                r'((?:Dipoles *\n|Dynamic.+?Dielectric|Dyson|Bare local|Local Exchange)[\s\S]+?\n *\[\d+\])',
                repeats=True, sub_parser=TextParser(quantities=module_quantities)
            )
        ]


class NetCDFParser(FileParser):
    def __init__(self):
        super().__init__()

    def init_parameters(self):
        self._keys = []

    @property
    def netcdf_file(self):
        if self._file_handler is None:
            try:
                self._file_handler = Dataset(self.mainfile)
            except Exception:
                self.logger.warn('Error loading file.')

        return self._file_handler

    def parse(self, key=None):
        self._results = dict() if self._results is None else self._results
        if self.netcdf_file is None:
            return

        self._keys = list(self.netcdf_file.variables.keys())
        for key in self._keys:
            self._results[key] = self.netcdf_file.variables[key][:].data


class InputParser(TextParser):
    def __init__(self):
        super().__init__()

    def init_quantities(self):
        def str_to_key_block(val_in):
            val = val_in.strip().split('\n')
            return val[0].strip(), [np.array(
                v.split('#')[0].split('|')[:-1], dtype=np.float64) for v in val[1:]]

        self._quantities = [
            Quantity('key_value', r'\n *(\w+) *= *(.+) *#*', repeats=True),
            Quantity('key_block', r'\n *\% *(\w+)([\s\S]+?)\%', repeats=True, str_operation=str_to_key_block)]


class YamboParser:
    def __init__(self):
        self.mainfile_parser = MainfileParser()
        self.input_parser = InputParser()
        self.netcdf_parser = NetCDFParser()
        self.metainfo_map = {
            'cpu': 'cores', 'threads': 'threads_per_core', 'threads_tot': 'threads_total',
            'io_nodes': 'nodes_io'
        }
        self._module = None

    def init_parser(self):
        self.mainfile_parser.mainfile = self.filepath
        self._module = None

    def parse_method(self, source, target=None):
        if source is None:
            return

        method = self.archive.run[-1].m_create(Method) if target is None else target

        section = method.m_create(self._module)

        for input in source.get('input', []):
            parameters = {key.strip(): val for key, val in input.get('key_value', [])}
            x_yambo_input = section.m_create(x_yambo_io, x_yambo_module.x_yambo_input)
            x_yambo_input.x_yambo_file = input.file
            x_yambo_input.x_yambo_sn = input.sn
            x_yambo_input.x_yambo_parameters = parameters

        for output in source.get('output', []):
            parameters = {key.strip(): val for key, val in output.get('key_value', [])}
            x_yambo_output = section.m_create(x_yambo_io, x_yambo_module.x_yambo_output)
            x_yambo_output.x_yambo_file = output.file
            x_yambo_output.x_yambo_sn = output.sn
            x_yambo_output.x_yambo_parameters = parameters

        for key, val in source.items():
            if key.startswith('x_yambo'):
                setattr(section, key, val)

    def parse_calculation(self, source, target=None):
        if source is None:
            return

        calc = self.archive.run[-1].m_create(Calculation) if target is None else target
        valence_conduction = source.get('valence_conduction', [
            source.get('valence', 0.), source.get('conduction', 0.)
        ])
        calc.energy = Energy(
            fermi=source.get('fermi', 0.0),
            highest_occupied=valence_conduction[0], lowest_unoccupied=valence_conduction[1])

        states = source.states_summary
        if states is not None:
            calc.x_yambo_filled_bands = [int(s) for s in states[0]]
            calc.x_yambo_empty_bands = [int(s) for s in states[-1]]

        if self.netcdf_parser.EIGENVALUES is not None:
            eigenvalues = calc.m_create(BandEnergies)
            eigenvalues.kpoints = np.transpose(self.netcdf_parser.get('K-POINTS'))
            eigenvalues.energies = self.netcdf_parser.EIGENVALUES * ureg.eV

        elif source.eigenenergies is not None:
            eigenvalues = calc.m_create(BandEnergies)
            eigenvalues.kpoints = source.eigenenergies.kpoints
            eigenvalues.kpoints_weights = source.eigenenergies.kpoints_weights
            # TODO deal with spin polarized data
            energies = source.eigenenergies.energies
            eigenvalues.energies = np.reshape(energies, (1, len(
                eigenvalues.kpoints), np.size(energies) // len(eigenvalues.kpoints))) * ureg.eV

        if self.netcdf_parser.QP_E_Eo_Z is not None or self.netcdf_parser.QP_E is not None:
            gw_band_energies = calc.m_create(BandEnergies)
            n_spin = self.netcdf_parser.QP_table.shape[1] // 2
            gw_band_energies.kpoints = self.netcdf_parser.QP_kpts.T
            if self.netcdf_parser.QP_E_Eo_Z is not None:
                qp_energy, bare_energy, z = self.netcdf_parser.QP_E_Eo_Z[0].T
            else:
                qp_energy = self.netcdf_parser.QP_E.T[0]
                bare_energy = self.netcdf_parser.QP_Eo
                z = self.netcdf_parser.QP_Z.T[0]
            # TODO verify if indeed energies are only for one kpoint
            shape = (n_spin, 1, len(qp_energy) // n_spin)
            gw_band_energies.value_qp = np.reshape(qp_energy, shape) * ureg.hartree
            gw_band_energies.value_ks = np.reshape(bare_energy, shape) * ureg.hartree
            gw_band_energies.qp_linearization_prefactor = np.reshape(z, shape)

        elif source.qp_energy is not None:
            gw_band_energies = calc.m_create(BandEnergies)
            qp_energy = source.qp_energy
            gw_band_energies.kpoints = [q.kpoint for q in qp_energy]
            energies = np.transpose([q.band for q in qp_energy])
            qp_energy = energies[2].T
            gw_band_energies.value_qp = np.reshape(qp_energy, (1, *np.shape(qp_energy))) * ureg.eV
            gw_band_energies.value_ks = np.reshape(energies[1].T, (1, *np.shape(qp_energy))) * ureg.eV
            gw_band_energies.qp_linearization_prefactor = np.reshape(energies[4].T, (1, *np.shape(qp_energy)))

        if self.netcdf_parser.Sx_Vxc is not None or self.netcdf_parser.Sx is not None:
            n_spin = self.netcdf_parser.QP_table.shape[1] // 2
            gw_band_energies = calc.eigenvalues[-1] if calc.eigenvalues else calc.m_create(BandEnergies)
            if self.netcdf_parser.Sx_Vxc is not None:
                if self.netcdf_parser.Sx_Vxc.shape[0] % 8 == 0:
                    qp = self.netcdf_parser.Sx_Vxc.reshape(-1, 8).T
                    sx, vxc = qp[4], qp[6]
                else:
                    qp = self.netcdf_parser.Sx_Vxc.reshape(-1, 7).T
                    sx, vxc = qp[3], qp[5]
            else:
                sx = self.netcdf_parser.Sx.T[0]
                vxc = self.netcdf_parser.Vxc.T[0]
            shape = (n_spin, 1, len(sx) // n_spin)
            gw_band_energies.value_exchange = np.reshape(sx, shape)
            gw_band_energies.value_xc_potential = np.reshape(vxc, shape)

        for key, val in source.items():
            if key.startswith('x_yambo') and val is not None:
                setattr(calc, key, val)

        return calc

    def parse_input(self):
        if self.mainfile_parser.cpu_files_io.input is None:
            return

        run = self.archive.run[-1]

        # read structure data from netcdf input file
        self.netcdf_parser.mainfile = os.path.join(self.maindir, self.mainfile_parser.cpu_files_io.input.file)
        if self.netcdf_parser.mainfile is not None:
            system = run.m_create(System)
            positions = self.netcdf_parser.get('ATOM_POS', [])
            n_atoms = self.netcdf_parser.N_ATOMS
            atom_numbers = np.hstack([[self.netcdf_parser.atomic_numbers[int(n)]] * int(
                n_atoms[int(n)]) for n in range(len(n_atoms))])
            system.atoms = Atoms(
                positions=np.reshape(positions, (np.size(positions) // 3, 3)) * ureg.bohr,
                labels=[chemical_symbols[int(n)] for n in atom_numbers])
            if self.netcdf_parser.LATTICE_VECTORS is not None:
                system.atoms.lattice_vectors = self.netcdf_parser.LATTICE_VECTORS * ureg.bohr

        # reference calculation
        energies_occupations = self.mainfile_parser.get('core_variables_setup', {}).get('energies_occupations')
        self.parse_calculation(energies_occupations)

        # input parameters from mainfile
        input = run.m_create(x_yambo_io)
        input.x_yambo_file = self.mainfile_parser.cpu_files_io.input.file
        input.x_yambo_sn = self.mainfile_parser.cpu_files_io.input.sn
        parameters = {key.strip(): val for key, val in self.mainfile_parser.cpu_files_io.input.get('key_value', [])}
        input.x_yambo_parameters = parameters

    def parse_dynamic_dielectric_matrix(self, module):
        source = module.dynamic_dielectric_matrix
        if source is None:
            return

        self._module = x_yambo_dynamic_dielectric_matrix
        self.parse_method(source)

        for output in source.get('output', []):
            path = os.path.join(self.maindir, os.path.dirname(output.get('file', '')))
            if not os.path.isdir(path):
                continue
            ddm = self.archive.run[-1].method[-1].x_yambo_dynamic_dielectric_matrix[-1]
            for filename in os.listdir(path):
                if 'pp_fragment_' not in filename:
                    continue
                self.netcdf_parser.mainfile = os.path.join(path, filename)
                if self.netcdf_parser.mainfile is None:
                    continue
                fragment = ddm.m_create(x_yambo_dynamic_dielectric_matrix_fragment)
                self.netcdf_parser.parse()
                for key in self.netcdf_parser._keys:
                    val = self.netcdf_parser.get(key)
                    if key.startswith('FREQ_PARS_sec_iq'):
                        fragment.x_yambo_FREQ_PARS_sec_iq = val
                    elif key.startswith('FREQ_sec_iq'):
                        fragment.x_yambo_FREQ_sec_iq = val
                    elif key.startswith('X_Q'):
                        fragment.x_yambo_X_Q = val

    def parse_local_xc_nonlocal_fock(self, module):
        source = module.local_xc_nonlocal_fock
        if source is None:
            return

        self._module = x_yambo_local_xc_nonlocal_fock
        self.parse_method(source)
        calc = self.parse_calculation(source.hf_occupations)

        if source.corrections is not None:
            calc = self.archive.run[-1].m_create(Calculation) if calc is None else calc
            if len(calc.x_yambo_local_xc_nonlocal_fock_bandenergies) == 0:
                band_energy = calc.m_create(x_yambo_local_xc_nonlocal_fock_bandenergies)
                sx, vxc = np.transpose([c.band for c in source.corrections])
                band_energy.x_yambo_sx = np.reshape(sx, (1, *np.shape(sx.T))) * ureg.eV
                band_energy.x_yambo_vxc = np.reshape(vxc, (1, *np.shape(vxc.T))) * ureg.eV

        if source.energy_xc is not None:
            calc.energy = Energy(xc=EnergyEntry(value=source.energy_xc))

    def parse_bare_xc(self, module):
        source = module.bare_xc
        if source is None:
            return

        self._module = x_yambo_bare_xc
        self.parse_method(source)
        calc = self.parse_calculation(source.hf_occupations)
        if source.xc_hf_dft is not None:
            calc = self.archive.run[-1].m_create(Calculation) if calc is None else calc
            if len(calc.x_yambo_bare_xc_bandenergies) == 0:
                band_energy = calc.m_create(x_yambo_bare_xc_bandenergies)
                hf, dft = np.transpose([h.band for h in source.xc_hf_dft])
                hf, dft = hf.T, dft.T
                band_energy.x_yambo_hf = np.reshape(hf, (1, *np.shape(hf))) * ureg.eV
                band_energy.x_yambo_dft = np.reshape(dft, (1, *np.shape(dft))) * ureg.eV

    def parse_dipoles(self, module):
        source = module.dipoles
        if source is None:
            return
        self._module = x_yambo_dipoles
        self.parse_method(source)

    def parse_dyson(self, module):
        source = module.dyson
        if source is None:
            return

        self._module = x_yambo_dyson

        if source.g0w0 is not None:
            self.parse_method(source.g0w0)

        if source.qp_properties is not None:
            self.parse_method(source.qp_properties, self.archive.run[-1].method[-1])
            for output in source.qp_properties.get('output', []):
                self.netcdf_parser.mainfile = os.path.join(self.maindir, output.get('file', ''))
                self.netcdf_parser.parse()
                self.parse_calculation(source.qp_properties)

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging

        self.init_parser()

        run = self.archive.m_create(Run)
        run.program = Program(
            name='YAMBO', version=self.mainfile_parser.get('version', ''),
            x_yambo_build=self.mainfile_parser.get('build', ''),
            x_yambo_hash=self.mainfile_parser.get('hash', ''))
        date = datetime.strptime(self.mainfile_parser.get(
            'date_start', '01/01/1970 00:00'), '%d/%m/%Y %H:%M') - datetime(1970, 1, 1)
        run.time_run = TimeRun(date_start=date.total_seconds())

        if self.mainfile_parser.cpu_files_io is not None:
            self.parse_input()
            for key, val in self.mainfile_parser.cpu_files_io.parameters:
                key = key.strip().replace('/', '').replace(' ', '_').lower()
                val = val == 'yes' if val in ['yes', 'no'] else val
                setattr(run, 'x_yambo_%s' % key, val)

        def parse_module(module):
            self.parse_dipoles(module)
            self.parse_dynamic_dielectric_matrix(module)
            self.parse_local_xc_nonlocal_fock(module)
            self.parse_bare_xc(module)
            self.parse_dyson(module)

        if self.mainfile_parser.transferred_momenta is not None:
            self._module = x_yambo_transferred_momenta
            self.parse_method(self.mainfile_parser.transferred_momenta)
            self.parse_calculation(self.mainfile_parser.transferred_momenta.qp_properties)
            for module in self.mainfile_parser.transferred_momenta.get('module', []):
                parse_module(module)

        for module in self.mainfile_parser.get('module', []):
            parse_module(module)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)
from nomad.datamodel.metainfo import simulation


m_package = Package()


class x_phonopy_input(MCategory):
    '''
    Information about properties that concern phonopy calculations.
    '''

    m_def = Category()


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_phonopy_displacement = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        Amplitude of the atom diplacement for the phonopy supercell
        ''',
        categories=[x_phonopy_input])

    x_phonopy_symprec = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        unit='meter',
        description='''
        Symmetry threshold for the space group identification of the crystal for which the
        vibrational properties are to be calculated
        ''',
        categories=[x_phonopy_input])


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_phonopy_original_system_ref = Quantity(
        type=simulation.system.System,
        shape=[],
        description='''
        Original cell from which the supercell for the DFT calculations was constructed
        ''')



"""parser doc

"""
#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import sys
import logging
import numpy as np
from nomad.datamodel import EntryArchive
from nomad.parsing import MatchingParser
# from . import metainfo  # pylint: disable=unused-import
from nexusparser.tools import nexus as read_nexus
from nexusparser.metainfo import nexus


def get_to_new_subsection(hdf_name, nxdef, nxdl_node, act_section):
    """hdf_name         name of the hdf group/field/attribute (None for definition)
    nxdef           application definition
    nxdl_node        node in the nxdl.xml
    act_class       actual class
    act_section     actual section in which the new entry needs to be picked up from
                    Note that if the new element did not exists, it is created now
    return          (new_class, new_section)
    TODO:   try to find also in the base section???

"""
    if hdf_name is None:
        nomad_def_name = 'nx_application_' + nxdef[2:]
        # nomad_class_name = nxdef
    elif nxdl_node.tag.endswith('field'):
        nxdl_f_a_name = nxdl_node.attrib['name'] if 'name' in nxdl_node.attrib else hdf_name
        nomad_def_name = 'nx_field_' + nxdl_f_a_name
        # nomad_class_name = self.get_nomad_classname(nxdl_f_a_name, None, "Field")
    elif nxdl_node.tag.endswith('group'):
        nxdl_g_name = nxdl_node.attrib['name'] \
            if 'name' in nxdl_node.attrib else nxdl_node.attrib['type'][2:].upper()
        nomad_def_name = 'nx_group_' + nxdl_g_name
        # nomad_class_name = self.get_nomad_classname(read_nexus.get_node_name(nxdl_node),
        #                                             nxdl_node.attrib['type'], "Group")
    else:
        nxdl_f_a_name = nxdl_node.attrib['name'] if 'name' in nxdl_node.attrib else hdf_name
        nomad_def_name = 'nx_attribute_' + nxdl_f_a_name
        # nomad_class_name = self.get_nomad_classname(nxdl_f_a_name, None, "Attribute")

    new_def = act_section.m_def.all_sub_sections[nomad_def_name]
    new_class = new_def.section_def.section_cls
    new_section = None
    for section in act_section.m_get_sub_sections(new_def):
        if hdf_name is None or (getattr(section, "nx_name") and section.nx_name == hdf_name):
            new_section = section
            break
    if new_section is None:
        act_section.m_create(new_class)
        new_section = act_section.m_get_sub_section(new_def, -1)
        if hdf_name is not None:
            new_section.nx_name = hdf_name
    return (new_class, new_section)


def get_value(hdf_value):
    """Get value from hdl5 node

"""
    if str(hdf_value.dtype) == 'bool':
        val = bool(hdf_value)
    elif hdf_value.dtype.kind in 'iufc':
        val = hdf_value
    else:
        val = str(hdf_value.astype(str))
    return val


def helper_nexus_populate(nxdl_attribute, act_section, val):
    """Handle info of units attribute, raise error if default or something else is found

"""
    try:
        if nxdl_attribute == "units":
            act_section.nx_unit = val[0]
        elif nxdl_attribute == "default":
            Exception("Quantity default' is not yet added by default to groups in Nomad schema")
    except Exception as exc:  # pylint: disable=broad-except
        print("Problem with storage!!!" + str(exc))


class NexusParser(MatchingParser):
    """NesusParser doc

"""
    def __init__(self):
        super().__init__(
            name='parsers/nexus', code_name='NEXUS', code_homepage='https://www.nexus.eu/',
            mainfile_mime_re=r'(application/.*)|(text/.*)',
            mainfile_name_re=(r'.*\.nxs'),
            supported_compressions=['gz', 'bz2', 'xz']
        )
        self.archive = None
        self.nxroot = None

#     def get_nomad_classname(self, xml_name, xml_type, suffix):
#         """Get nomad classname from xml file

# """
#         if suffix == 'Attribute' or suffix == 'Field' or xml_type[2:].upper() != xml_name:
#             name = xml_name + suffix
#         else:
#             name = xml_type + suffix
#         return name

    def nexus_populate(self, hdf_info, nxdef, nxdl_path, val):
        """Walks through hdf_namelist and generate nxdl nodes

"""
        hdf_path = hdf_info['hdf_path']
        hdf_node = hdf_info['hdf_node']
        print('%%%%%%%%%%%%%%')
        # print(nxdef+':'+'.'.join(p.getroottree().getpath(p) for p in nxdl_path)+
        # ' - '+val[0]+ ("..." if len(val) > 1 else ''))
        if nxdl_path is not None:
            print((nxdef or '???') + ':' + '.'.join(p if isinstance(p, str) else
                                                    read_nexus.get_node_name(p)
                                                    for p in nxdl_path) + ' - \
' + val[0] + ("..." if len(val) > 1 else ''))
            act_section = self.nxroot
            hdf_namelist = hdf_path.split('/')[1:]
            act_section = get_to_new_subsection(None, nxdef, None, act_section)[1]
            path_level = 1
            for hdf_name in hdf_namelist:
                nxdl_node = nxdl_path[path_level] if path_level < len(nxdl_path) else hdf_name
                act_section = get_to_new_subsection(hdf_name, nxdef,
                                                    nxdl_node, act_section)[1]
                path_level += 1
            if path_level < len(nxdl_path):
                nxdl_attribute = nxdl_path[path_level]
                if isinstance(nxdl_attribute, str):
                    # conventional attribute not in schema. Only necessary,
                    # if schema is not populated according
                    helper_nexus_populate(nxdl_attribute, act_section, val)
                else:
                    # attribute in schema
                    act_section = \
                        get_to_new_subsection(nxdl_attribute.attrib['name'], nxdef,
                                              nxdl_attribute, act_section)[1]
                    try:
                        act_section.nx_value = val[0]
                    except AttributeError as exc:
                        print("Problem with storage!!!" + str(exc))
                    except TypeError as exc:
                        print("Problem with storage!!!" + str(exc))
            else:
                try:
                    data_field = get_value(hdf_node[...])
                    if hdf_node[...].dtype.kind in 'iufc' and \
                            isinstance(data_field, np.ndarray) and \
                            data_field.size > 1:
                        data_field = np.array([
                            np.mean(data_field),
                            np.var(data_field),
                            np.min(data_field),
                            np.max(data_field)
                        ])
                    act_section.nx_value = data_field
                except TypeError as exc:
                    print("Problem with storage!!!" + str(exc))
        else:
            print('NOT IN SCHEMA - skipped')
        print('%%%%%%%%%%%%%%')

    def parse(self, mainfile: str, archive: EntryArchive, logger=None):
        stdout_handler = logging.StreamHandler(sys.stdout)
        stdout_handler.setLevel(logging.DEBUG)
        stdout_handler.setFormatter(logging.Formatter("%(levelname)s: %(message)s"))
        logger.addHandler(stdout_handler)

        self.archive = archive
        # TODO ask Markus S. whether to disable or not this pylint error
        self.archive.m_create(nexus.Nexus)  # pylint: disable=no-member
        self.nxroot = self.archive.nexus

        nexus_helper = read_nexus.HandleNexus(logger, [mainfile])
        nexus_helper.process_nexus_master_file(self.nexus_populate)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import List, Dict, Optional
from collections import defaultdict
from enum import Enum
from pydantic import BaseModel, Field
from fastapi import (
    APIRouter, Depends, Request
)
from elasticsearch_dsl import Search
from elasticsearch_dsl.utils import AttrList
from elasticsearch.exceptions import RequestError

from nomad.metainfo.elasticsearch_extension import entry_index, entry_type

from .auth import create_user_dependency
from ..models import User


router = APIRouter()


# This is a dynamically create enum class for enumerating all allowed
# quantities. FastAPI uses python enums to validate and document options.
class SuggestableQuantityEnum(str, Enum):
    pass


SuggestableQuantity = SuggestableQuantityEnum(  # type: ignore
    'SuggestableQuantity',
    {name: name for name in entry_type.suggestions.keys()}
)


class SuggestionError(Exception): pass


class Suggestion(BaseModel):
    value: str = Field(None, description='The returned suggestion.')
    weight: Optional[float] = Field(None, description='The suggestion weight.')


class SuggestionsRequest(BaseModel):
    quantities: List[SuggestableQuantity] = Field(  # type: ignore
        None,
        description='List of quantities for which the suggestions are retrieved.'
    )
    input: str = Field(
        None,
        description='The input that is used as a basis for returning a suggestion.'
    )


@router.post(
    '',
    tags=['suggestions'],
    summary='Get a list of suggestions for the given quantity names and input.',
    response_model=Dict[str, List[Suggestion]],
    response_model_exclude_unset=True,
    response_model_exclude_none=True)
async def get_suggestions(
        request: Request,
        data: SuggestionsRequest,
        user: User = Depends(create_user_dependency())):

    search = Search(index=entry_index.index_name)
    quantities: List[str] = data.quantities
    quantities_es = [x.replace(".", "-") for x in quantities]
    for quantity, quantity_es in zip(quantities, quantities_es):
        annotation = entry_type.suggestions[quantity]
        postfix = ".suggestion" if annotation.field else "__suggestion"
        search = search.suggest(quantity_es, data.input, completion={
            'field': f'{quantity}{postfix}',
            'size': 5,
            'skip_duplicates': True,
            # 'fuzzy': {},
        })
    search = search.extra(_source=quantities)

    try:
        # For some reason calling the search.extra()-method messes up the type
        # information for the Search-object. This is why linting is disabled
        # here.
        es_response = search.execute()  # pylint: disable=no-member
    except RequestError as e:
        raise SuggestionError(e)

    # We return the original field in the source document.
    response: Dict[str, List[Suggestion]] = defaultdict(list)
    suggestions: Dict[str, Dict[str, float]] = defaultdict(dict)

    def add_suggestion(quantity, value, weight):
        values = suggestions[quantity]
        if value not in values or values[value] < weight:
            suggestions[quantity][value] = weight

    for quantity, quantity_es in zip(quantities, quantities_es):
        variants = entry_type.suggestions[quantity].variants
        for option in es_response.suggest[quantity_es][0].options:
            weight = option._score

            # We use the original input text to do the matching. This works
            # better than the text returned by the completion suggester
            # (option.text), since it can match several items if there are
            # multiple values per quantity.
            text = data.input.lower().strip()

            # Nested fields use the nested document as _source: we need to
            # modify the path accordingly.
            try:
                nested_field = option._nested.field
            except AttributeError:
                nested_field = None
            quantity_path = quantity[len(nested_field):] if nested_field else quantity

            # Gather all options recursively. The source may contain
            # inner/nested sections and also values may be lists.
            def gather_options(root, parts, options):
                original = root
                for i, part in enumerate(parts):
                    if part != "":
                        original = original[part]
                        if isinstance(original, AttrList):
                            for item in original:
                                gather_options(item, parts[i + 1:], options)
                            return
                original_type = type(original)
                if original_type == str:
                    options.append(original)
                elif original_type == list:
                    for item in original:
                        options.append(item)
            options: List[str] = []
            parts = quantity_path.split('.')
            gather_options(option._source, parts, options)

            # There may be multiple options and we have to look which options
            # were actually matched (the completion suggester does not have a
            # mechanism for saving which item in a list was matched.). If the
            # value has several variants, we have to expand the original source
            # value and return only the best match.
            for option in options:
                if variants:
                    best_match = float("Inf")
                    best_option = option
                    for variant in variants(option):
                        match_start = variant.lower().strip().find(text)
                        if match_start >= 0 and match_start < best_match:
                            best_match = match_start
                            best_option = variant
                    add_suggestion(quantity, best_option, weight)
                elif text in option.lower().strip():
                    add_suggestion(quantity, option, weight)

    for name, suggestion in suggestions.items():
        for value, weight in suggestion.items():
            response[name].append(Suggestion(value=value, weight=weight))

    return response



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
"""Helper functions commonly used by the convert routine."""

from typing import Tuple, Callable
import re
import xml.etree.ElementTree as ET

import numpy as np

from nexusparser.tools import nexus


def generate_template_from_nxdl(root, template, path=None, nxdl_root=None):
    """Helper function to generate a template dictionary for given NXDL"""
    if nxdl_root is None:
        nxdl_root = root
        root = get_first_group(root)
    if path is None:
        path = ""

    tag = remove_namespace_from_tag(root.tag)

    if tag == "doc":
        return

    suffix = ""
    if "name" in root.attrib:
        suffix = root.attrib['name']
    elif "type" in root.attrib:
        nexus_class = convert_nexus_to_caps(root.attrib['type'])
        hdf5name = f"[{convert_nexus_to_suggested_name(root.attrib['type'])}]"
        suffix = f"{nexus_class}{hdf5name}"

    if tag == "attribute":
        suffix = f"@{suffix}"

    path = path + "/" + suffix

    # Only add fields or attributes to the dictionary
    if tag in ("field", "attribute"):
        optionality = get_required_string(root)
        if optionality == "required":
            optional_parent = check_for_optional_parent(path, nxdl_root)
            optionality = "required" if optional_parent == "<<NOT_FOUND>>" else "optional"
            if optional_parent != "<<NOT_FOUND>>":
                template.optional_parents.append(optional_parent)
        template[optionality][path] = None

        # Only add units if it is a field and the the units are defined but not set to NX_UNITLESS
        if tag == "field" \
           and ("units" in root.attrib.keys() and root.attrib["units"] != "NX_UNITLESS"):
            template[optionality][f"{path}/@units"] = None

    for child in root:
        generate_template_from_nxdl(child, template, path, nxdl_root)


def get_required_string(elem):
    """Helper function to return nicely formatted names for optionality."""
    return nexus.get_required_string(elem)[2:-2].lower()


def convert_nexus_to_caps(nexus_name):
    """Helper function to convert a Nexus class from <NxClass> to <CLASS>."""
    return nexus_name[2:].upper()


def convert_nexus_to_suggested_name(nexus_name):
    """Helper function to suggest a name for a group from its Nexus class."""
    return nexus_name[2:]


def convert_data_converter_entry_to_nxdl_path_entry(entry) -> str:
    """
    Helper function to convert data converter style entry to NXDL style entry:
    ENTRY[entry] -> ENTRY
    """
    regex = re.compile(r'(.*?)(?=\[)')
    results = regex.search(entry)
    return entry if results is None else results.group(1)


def convert_data_converter_dict_to_nxdl_path(path) -> str:
    """
    Helper function to convert data converter style path to NXDL style path:
    /ENTRY[entry]/sample -> /ENTRY/sample
    """
    nxdl_path = ''
    for entry in path.split('/')[1:]:
        nxdl_path += '/' + convert_data_converter_entry_to_nxdl_path_entry(entry)
    return nxdl_path


def get_name_from_data_dict_entry(entry) -> str:
    """Helper function to get entry name from data converter style entry

    ENTRY[entry] -> entry
    """
    regex = re.compile(r'(?<=\[)(.*?)(?=\])')
    results = regex.search(entry)
    return entry if results is None else results.group(1)


def convert_data_dict_path_to_hdf5_path(path) -> str:
    """Helper function to convert data converter style path to HDF5 style path

    /ENTRY[entry]/sample -> /entry/sample
    """
    hdf5path = ''
    for entry in path.split('/')[1:]:
        hdf5path += '/' + get_name_from_data_dict_entry(entry)
    return hdf5path


def is_value_valid_element_of_enum(value, elem) -> Tuple[bool, list]:
    """Checks whether a value has to be specific from the NXDL enumeration and returns options."""
    if elem is not None:
        has_enums, enums = nexus.get_enums(elem)
        if has_enums and (isinstance(value, list) or value not in enums[0:-1] or value == ""):
            return False, enums
    return True, []


NUMPY_FLOAT_TYPES = (np.half, np.float16, np.single, np.double, np.longdouble)
NUMPY_INT_TYPES = (np.short, np.intc, np.int_)
NUMPY_UINT_TYPES = (np.ushort, np.uintc, np.uint)

NEXUS_TO_PYTHON_DATA_TYPES = {
    "ISO8601": (str),
    "NX_BINARY": (bytes, bytearray, np.byte, np.ubyte, np.ndarray),
    "NX_BOOLEAN": (bool, np.ndarray, np.bool_),
    "NX_CHAR": (str, np.ndarray, np.chararray),
    "NX_DATE_TIME": (str),
    "NX_FLOAT": (float, np.ndarray, np.floating),
    "NX_INT": (int, np.ndarray, np.signedinteger),
    "NX_UINT": (np.ndarray, np.unsignedinteger),
    "NX_NUMBER": (int, float, np.ndarray, np.signedinteger, np.unsignedinteger, np.floating),
    "NX_POSINT": (int, np.ndarray, np.signedinteger),  # > 0 is checked in is_valid_data_field()
    "NXDL_TYPE_UNAVAILABLE": (str)  # Defaults to a string if a type is not provided.
}


def check_all_children_for_callable(objects: list, check: Callable, *args) -> bool:
    """Checks whether all objects in list are validated by given callable."""
    for obj in objects:
        if not check(obj, *args):
            return False

    return True


def is_valid_data_type(value, accepted_types):
    """Checks whether the given value or its children are of an accepted type."""
    if not isinstance(value, list):
        return isinstance(value, accepted_types)

    return check_all_children_for_callable(value, isinstance, accepted_types)


def is_positive_int(value):
    """Checks whether the given value or its children are positive."""
    is_greater_than = lambda x: x.flat[0] > 0 if isinstance(x, np.ndarray) else x > 0

    if isinstance(value, list):
        return check_all_children_for_callable(value, is_greater_than)

    return value.flat[0] > 0 if isinstance(value, np.ndarray) else value > 0


def is_valid_data_field(value, nxdl_type, path):
    """Checks whether a given value is valid according to what is defined in the NXDL."""
    accepted_types = NEXUS_TO_PYTHON_DATA_TYPES[nxdl_type]

    if not is_valid_data_type(value, accepted_types):
        raise Exception(f"The value at {path} should be of Python type: {accepted_types}"
                        f", as defined in the NXDL as {nxdl_type}.")

    if nxdl_type == "NX_POSINT" and not is_positive_int(value):
        raise Exception(f"The value at {path} should be a positive int.")

    if nxdl_type in ("ISO8601", "NX_DATE_TIME"):
        iso8601 = re.compile(r"^(\d{4})-(\d{2})-(\d{2})T(\d{2}):(\d{2}):(\d{2}(?:"
                             r"\.\d*)?)(((?!-00:00)(\+|-)(\d{2}):(\d{2})|Z){1})$")
        results = iso8601.search(value)
        if results is None:
            raise Exception(f"The date at {path} should be a timezone aware ISO8601 "
                            f"formatted str. For example, 2022-01-22T12:14:12.05018Z"
                            f" or 2022-01-22T12:14:12.05018+00:00.")


def path_in_data_dict(nxdl_path: str, data: dict) -> Tuple[bool, str]:
    """Checks if there is an accepted variation of path in the dictionary & returns the path."""
    for key in data.keys():
        if nxdl_path == convert_data_converter_dict_to_nxdl_path(key):
            return True, key
    return False, ""


def check_for_optional_parent(path: str, nxdl_root: ET.Element) -> str:
    """Finds a parent in the branch that is optional and returns it's path or s<<NOT_FOUND>>."""
    parent_path = path.rsplit("/", 1)[0]

    if parent_path == "":
        return "<<NOT_FOUND>>"

    parent_nxdl_path = convert_data_converter_dict_to_nxdl_path(parent_path)
    elem = nexus.get_node_at_nxdl_path(nxdl_path=parent_nxdl_path, elem=nxdl_root)

    if nexus.get_required_string(elem) in ("<<OPTIONAL>>", "<<RECOMMENDED>>"):
        return parent_path

    return check_for_optional_parent(parent_path, nxdl_root)


def is_node_required(nxdl_key, nxdl_root):
    """Checks whether a node at given nxdl path is required"""
    if nxdl_key[nxdl_key.rindex("/") + 1:] == "@units":
        return False
    if nxdl_key[nxdl_key.rindex("/") + 1] == "@":
        nxdl_key = nxdl_key[0:nxdl_key.rindex("/") + 1] + nxdl_key[nxdl_key.rindex("/") + 2:]
    node = nexus.get_node_at_nxdl_path(nxdl_key, elem=nxdl_root)
    return nexus.get_required_string(node) == "<<REQUIRED>>"


def all_required_children_are_set(optional_parent_path, data, nxdl_root):
    """Walks over optional parent's children and makes sure all required ones are set"""
    optional_parent_path = convert_data_converter_dict_to_nxdl_path(optional_parent_path)
    for key in data:
        nxdl_key = convert_data_converter_dict_to_nxdl_path(key)
        if optional_parent_path in nxdl_key \
           and is_node_required(nxdl_key, nxdl_root) \
           and data[key] is None:
            return False

    return True


def check_optionality_based_on_parent_group(
        path,
        nxdl_path,
        nxdl_root,
        data,
        template):
    """Checks whether field is part of an optional parent and then confirms its optionality"""
    for optional_parent in template["optional_parents"]:
        optional_parent_nxdl = convert_data_converter_dict_to_nxdl_path(optional_parent)
        if optional_parent_nxdl in nxdl_path \
           and not all_required_children_are_set(optional_parent, data, nxdl_root):
            raise Exception(f"The data entry, {path}, has an optional parent, "
                            f"{optional_parent}, with required children set. Either"
                            f" provide no children for {optional_parent} or provide"
                            f" all required ones.")


def ensure_all_required_fields_exist(template, data):
    """Checks whether all the required fields are in the returned data object."""
    for path in template["required"]:
        entry_name = get_name_from_data_dict_entry(path[path.rindex('/') + 1:])
        if entry_name == "@units":
            continue
        nxdl_path = convert_data_converter_dict_to_nxdl_path(path)
        is_path_in_data_dict, renamed_path = path_in_data_dict(nxdl_path, data)
        if not is_path_in_data_dict or data[renamed_path] is None:
            raise Exception(f"The data entry corresponding to {path} is required and"
                            f" hasn't been supplied by the reader.")


def validate_data_dict(template, data, nxdl_root: ET.Element):
    """Checks whether all the required paths from the template are returned in data dict."""
    assert nxdl_root is not None, "The NXDL file hasn't been loaded."

    # Make sure all required fields exist.
    ensure_all_required_fields_exist(template, data)

    for path in data.get_documented().keys():
        if data[path] is not None:
            entry_name = get_name_from_data_dict_entry(path[path.rindex('/') + 1:])
            nxdl_path = convert_data_converter_dict_to_nxdl_path(path)

            if entry_name == "@units":
                continue
            elif entry_name[0] == "@":
                index_of_at = nxdl_path.rindex("@")
                nxdl_path = nxdl_path[0:index_of_at] + nxdl_path[index_of_at + 1:]

            elem = nexus.get_node_at_nxdl_path(nxdl_path=nxdl_path, elem=nxdl_root)

            # Only check for validation in the NXDL if we did find the entry
            # otherwise we just pass it along
            if elem is not None \
               and elem.attrib["name"] == entry_name \
               and remove_namespace_from_tag(elem.tag) in ("field", "attribute"):
                check_optionality_based_on_parent_group(path, nxdl_path, nxdl_root, data, template)

                attrib = elem.attrib
                nxdl_type = attrib["type"] if "type" in attrib.keys() else "NXDL_TYPE_UNAVAILABLE"
                is_valid_data_field(data[path], nxdl_type, path)
                is_valid_enum, enums = is_value_valid_element_of_enum(data[path], elem)
                if not is_valid_enum:
                    raise Exception(f"The value at {path} should be"
                                    f" one of the following strings: {enums}")

    return True


def remove_namespace_from_tag(tag):
    """Helper function to remove the namespace from an XML tag."""
    return tag.split("}")[-1]


def get_first_group(root):
    """Helper function to get the actual first group element from the NXDL."""
    for child in root:
        if remove_namespace_from_tag(child.tag) == "group":
            return child
    return root



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import numpy as np            # pylint: disable=unused-import
import typing                 # pylint: disable=unused-import
from nomad.metainfo import (  # pylint: disable=unused-import
    MSection, MCategory, Category, Package, Quantity, Section, SubSection, SectionProxy,
    Reference
)

from nomad.datamodel.metainfo import simulation


m_package = Package()


class CalcTypes(MSection):

    m_def = Section(validate=False)

    x_mp_label = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_value = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class Origins(MSection):

    m_def = Section(validate=False)

    x_mp_name = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_task_id = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_last_updated = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class Run(simulation.run.Run):

    m_def = Section(validate=False, extends_base_section=True)

    x_mp_emmet_version = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_pymatgen_version = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_build_date = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_cif = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_n_tags = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_mp_tags = Quantity(
        type=str,
        shape=['x_mp_n_tags'],
        description='''
        ''')

    x_mp_material_id = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_deprecated = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_n_tasks = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_mp_task_ids = Quantity(
        type=str,
        shape=['x_mp_n_tasks'],
        description='''
        ''')

    x_mp_icsd_id = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_mp_icsd_ids = Quantity(
        type=np.dtype(np.int32),
        shape=['x_mp_n_tasks'],
        description='''
        ''')

    x_mp_last_updated = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_created_at = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_calc_types = SubSection(sub_section=CalcTypes.m_def, repeats=True)

    x_mp_origins = SubSection(sub_section=Origins.m_def, repeats=True)


class Composition(MSection):

    m_def = Section(validate=False)

    x_mp_label = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_value = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')


class Symmetry(MSection):

    m_def = Section(validate=False)

    x_mp_symprec = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_mp_version = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_source = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_symbol = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_number = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_mp_point_group = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_crystal_system = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_hall = Quantity(
        type=str,
        shape=[],
        description='''
        ''')


class System(simulation.system.System):

    m_def = Section(validate=False, extends_base_section=True)

    x_mp_formula_anonymous = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_oxide_type = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_chemsys = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_formula_pretty = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_volume = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_mp_density = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_mp_density_atomic = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')

    x_mp_nelements = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_mp_elements = Quantity(
        type=str,
        shape=['x_mp_nelements'],
        description='''
        ''')

    x_mp_nsites = Quantity(
        type=np.dtype(np.int32),
        shape=[],
        description='''
        ''')

    x_mp_symmetry = SubSection(sub_section=Symmetry.m_def, repeats=True)

    x_mp_composition = SubSection(sub_section=Composition.m_def, repeats=True)

    x_mp_composition_reduced = SubSection(sub_section=Composition.m_def, repeats=True)


class Hubbard(MSection):

    m_def = Section(validate=False)

    x_mp_element = Quantity(
        type=str,
        shape=[],
        description='''
        ''')

    x_mp_hubbard = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')


class Method(simulation.method.Method):

    m_def = Section(validate=False, extends_base_section=True)

    x_mp_is_compatible = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_mp_is_hubbard = Quantity(
        type=bool,
        shape=[],
        description='''
        ''')

    x_mp_hubbards = SubSection(sub_section=Hubbard.m_def, repeats=True)


class Calculation(simulation.calculation.Calculation):

    m_def = Section(validate=False, extends_base_section=True)

    x_mp_uncorrected_energy_per_atom = Quantity(
        type=np.dtype(np.float64),
        shape=[],
        description='''
        ''')



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Union
from fastapi import APIRouter, Query, Path, HTTPException, status, Depends
from datetime import datetime, date
from elasticsearch_dsl import Q

from nomad import utils
from nomad.utils import strip
from nomad.search import search
from nomad.app.v1.models import MetadataPagination, HTTPExceptionModel
from nomad.app.v1.utils import create_responses

from ..common import rdf_response
from ..mapping import Mapping

router = APIRouter()
default_tag = 'dcat'

logger = utils.get_logger(__name__)


_bad_id_response = status.HTTP_404_NOT_FOUND, {
    'model': HTTPExceptionModel,
    'description': strip('''
        Dataset not found. The given id does not match any dataset.''')}

_raw_response = 200, {
    'content': {'application/octet': {}},
    'description': 'The response. The returned content type depends on the format parameter.'}


@router.get(
    '/datasets/{entry_id}', tags=[default_tag],
    summary='Returns a DCAT dataset for a given NOMAD entry id.',
    responses=create_responses(_bad_id_response, _raw_response)
)
async def get_dataset(
    entry_id: str = Path(..., description='The unique NOMAD entry id.'),
    rdf_respose=Depends(rdf_response)
):
    ''' Returns a DCAT dataset for a given NOMAD entry id. '''

    results = search(owner='public', query=dict(entry_id=entry_id))
    if results.pagination.total == 0:
        raise HTTPException(
            status_code=_bad_id_response[0],
            detail=_bad_id_response[1]['description'])

    entry = results.data[0]

    mapping = Mapping()
    mapping.map_entry(entry)
    return rdf_respose(mapping.g)


@router.get(
    '/catalog/', tags=[default_tag],
    summary='Returns a DCAT dataset for a given NOMAD entry id.',
    responses=create_responses(_raw_response)
)
async def get_catalog(
    after: str = Query(None, description='return entries after the given entry_id'),
    modified_since: Union[datetime, date] = Query(None, description='maximum entry time (e.g. upload time)'),
    rdf_respose=Depends(rdf_response)
):
    ''' Returns a DCAT dataset for a given NOMAD entry id. '''

    search_query = Q()
    if modified_since is not None:
        modified_clause = Q('range', upload_create_time=dict(gte=modified_since))
        modified_clause |= Q('range', last_edit_time=dict(gte=modified_since))
        modified_clause |= Q('range', last_processing_time=dict(gte=modified_since))
        search_query &= modified_clause

    pagination = MetadataPagination(page_after_value=after)

    search_response = search(owner='public', query=search_query, pagination=pagination)

    mapping = Mapping()
    mapping.map_catalog(
        search_response.data,
        total=search_response.pagination.total,
        after=after,
        modified_since=modified_since, slim=False)
    return rdf_respose(mapping.g)



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD. See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from typing import Optional
from fastapi import Response, Query, Header
import urllib.parse
from rdflib import Graph
from enum import Enum

from nomad import config


root_path = f'{config.services.api_base_path}/dcat'
base_url = config.api_url(api='dcat')


def url(*args, **kwargs):
    ''' Returns the full dcat api url for the given path (args) and query (kwargs) parameters. '''
    url = f'{base_url.rstrip("/")}/{"/".join(args).lstrip("/")}'

    if len(kwargs) > 0:
        return f'{url}?{urllib.parse.urlencode(kwargs)}'
    else:
        return url


class Formats(str, Enum):
    xml = 'xml',
    n3 = 'n3',
    turtle = 'turtle',
    nt = 'nt',
    pretty_xml = 'pretty-xml',
    trig = 'trig'


all_repsonse_types = {
    'application/xml': 'xml',
    'application/rdf+prettyxml': 'pretty-xml',
    'application/rdf': 'xml',
    'application/rdf+xml': 'xml',
    'text/plain': 'n3',
    'text/turtle': 'turtle',
    'text/nt': 'nt',
    'text/n3': 'n3',
    'text/rdf+n3': 'n3',
    'text/rdf+nt': 'nt',
    'text/rdf+turtle': 'turtle',
    'application/x-trig': 'trig'
}

response_types = [
    'application/xml',
    'application/rdf+xml',
    'application/rdf+pretty-xml',
    'text/plain',
    'text/turtle',
    'text/rdf+n3',
    'text/rdf+nt',
    'application/x-trig']


def rdf_response(
    format: Optional[Formats] = Query(None), accept: Optional[str] = Header(None)
):
    format_ = format.value if format else None
    if format_ is None:
        if accept:
            format_ = all_repsonse_types.get(accept, 'pretty-xml')
        else:
            format_ = 'pretty-xml'

    def create_response(g: Graph):
        try:
            content_type = next(key for key, value in all_repsonse_types.items() if value == format_)
        except StopIteration:
            content_type = 'application/xml' if format_ in ['xml', 'pretty-xml'] else f'text/format'

        return Response(
            g.serialize(format=format_).decode('utf-8'), media_type=content_type)

    return create_response



#
# Copyright The NOMAD Authors.
#
# This file is part of NOMAD.
# See https://nomad-lab.eu for further info.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import os
import logging
from datetime import datetime
import numpy as np

from nomad.units import ureg
from nomad.parsing.file_parser import TextParser, Quantity
from nomad.datamodel.metainfo.simulation.run import Run, Program, TimeRun
from nomad.datamodel.metainfo.simulation.system import System, Atoms
from nomad.datamodel.metainfo.simulation.method import (
    Electronic, Method, BasisSet, BasisSetAtomCentered, Scf as ScfMethod, DFT,
    XCFunctional, Functional)
from nomad.datamodel.metainfo.simulation.calculation import (
    Calculation, Energy, EnergyEntry, Forces, ForcesEntry, ScfIteration, BandEnergies,
    Multipoles, MultipolesEntry, Charges, ChargesValue)
from nomad.datamodel.metainfo.workflow import Workflow, GeometryOptimization
from .metainfo.psi4 import x_psi4_root_information
from .metainfo import m_env


class OutParser(TextParser):
    def __init__(self):
        super().__init__(None)

    def init_quantities(self):
        re_f = r'-*\d+\.\d+[Ee]*[-+]*\d*'

        def str_to_basis(val_in):
            data = dict()
            for line in val_in.strip().split('\n'):
                line = line.split(':')
                if len(line) == 2:
                    data[line[0].strip()] = line[1].strip()
            return data

        def str_to_orbital_energies(val_in):
            val = val_in.strip().split()
            orbitals = [val[i] for i in range(0, len(val), 2)]
            energies = [float(val[i]) for i in range(1, len(val), 2)]
            return orbitals, energies

        def str_to_eigenvalues(val_in):
            orbital_energies = []
            for val in val_in.strip().split('\n'):
                if 'occupied orbitals' in val:
                    orbital_energies.append([[], []])
                elif 'Active orbitals' in val:
                    pass
                else:
                    val = val.strip().split()
                    if len(val) % 3 != 0:
                        continue
                    orbital_energies[-1][0].extend([val[i] for i in range(2, len(val), 3)])
                    orbital_energies[-1][1].extend([float(val[i]) for i in range(1, len(val), 3)])
            return orbital_energies

        def str_to_parameters(val_in):
            val_in = val_in.strip().split('\n')
            name = val_in[0].split()[0]
            parameters = dict()
            separator = '='
            if '=>' in val_in[-1]:
                separator = '=>'
            elif ':' in val_in[-1]:
                separator = ':'
            for val in val_in:
                val = [v.rstrip('!').strip() for v in val.split(separator)]
                if len(val) == 2:
                    try:
                        val[1] = float(val[1])
                    except Exception:
                        pass
                    if val[1] in ['TRUE', 'FALSE']:
                        val[1] = val[1] == 'TRUE'
                    parameters[val[0]] = val[1]
            return name, parameters

        charges_quantities = [
            Quantity(
                'atom',
                rf'\d+ +[A-Z][a-z]* +({re_f}) +({re_f}) +({re_f}) +({re_f})',
                repeats=True, dtype=np.dtype(np.float64)),
            Quantity(
                'total',
                rf'Total alpha = +({re_f}), Total beta = +({re_f}), Total charge = +({re_f})',
                dtype=np.dtype(np.float64))]

        properties_quantities = [
            Quantity(
                'multipole_moments',
                r'Multipole +Electric \(a\.u\.\) .+\s+\-+([\s\S]+?)\-{10}',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'dipole',
                        r'Dipole [XYZ] +: +(.+)',
                        repeats=True, dtype=np.dtype(np.float64)),
                    Quantity(
                        'quadrupole',
                        r'Quadrupole [XYZ]{2} +: +(.+)',
                        repeats=True, dtype=np.dtype(np.float64)),
                    Quantity(
                        'octupole',
                        r'Octupole [XYZ]{3} +: +(.+)',
                        repeats=True, dtype=np.dtype(np.float64)),
                    Quantity(
                        'hexadecapole',
                        r'Hexadecapole [XYZ]{4} +: +(.+)',
                        repeats=True, dtype=np.dtype(np.float64)),
                    Quantity(
                        'npole',
                        r'32-pole [XYZ]{5} +: +(.+)',
                        repeats=True, dtype=np.dtype(np.float64))
                ])
            ),
            Quantity(
                'nuclear_dipole_moment',
                rf'Nuclear Dipole Moment: \(a\.u\.\)\s+X: +({re_f}) +Y: +({re_f}) +Z: +({re_f})',
                dtype=np.dtype(np.float64)),
            Quantity(
                'electronic_dipole_moment',
                rf'Electronic Dipole Moment: \(a\.u\.\)\s+X: +({re_f}) +Y: +({re_f}) +Z: +({re_f})',
                dtype=np.dtype(np.float64)),
            Quantity(
                'total_dipole_moment',
                rf'Dipole Moment: \(a\.u\.\)\s+X: +({re_f}) +Y: +({re_f}) +Z: +({re_f})',
                dtype=np.dtype(np.float64)),
            Quantity(
                'mulliken_charges',
                r'Mulliken Charges: \(a\.u\.\)\s+.+([\s\S]+?Total charge.+)',
                sub_parser=TextParser(quantities=charges_quantities)),
            Quantity(
                'lowdin_charges',
                r'Lowdin Charges: \(a\.u\.\)\s+.+([\s\S]+?Total charge.+)',
                sub_parser=TextParser(quantities=charges_quantities))
        ]

        scf_quantities = [
            Quantity(
                'geometry',
                r'Geometry <==([\s\S]+?)==>', sub_parser=TextParser(quantities=[
                    Quantity(
                        'atoms',
                        rf'([A-Z][a-z]* +{re_f} +{re_f} +{re_f} +{re_f}\n)',
                        repeats=True),
                    Quantity('molecular_point_group', r'Molecular point group: *(\w+)', dtype=str),
                    Quantity('full_point_group', r'Full point group: *(\w+)', dtype=str),
                    Quantity('symmetry', r'Running in (\w+) symmetry\.', dtype=str),
                    Quantity(
                        'rotational_constants',
                        rf'Rotational constants: *A = *({re_f}) *B = *({re_f}) *C = *({re_f}) \[cm\^-1\]',
                        dtype=np.dtype(np.float64)),
                    Quantity('nuclear_repulsion', rf'Nuclear repulsion = *({re_f})', dtype=np.float64),
                    Quantity('charge', r'Charge *= (\d+)', dtype=np.float64),
                    Quantity('multiplicity', r'Multiplicity *= (\d+)', dtype=np.float64),
                    Quantity('electrons', r'Electrons *= (\d+)', dtype=np.float64),
                    Quantity('nalpha', r'Nalpha *= (\d+)', dtype=np.float64),
                    Quantity('nbeta', r'Nbeta *= (\d+)', dtype=np.float64)])),
            Quantity(
                'algorithm',
                r'Algorithm <==([\s\S]+?)==>', sub_parser=TextParser(quantities=[
                    Quantity(
                        'minimization_algorithm',
                        r'SCF Algorithm Type is (.+)\.', dtype=str),
                    Quantity(
                        'x_psi4_diis',
                        r'DIIS (.+)\.', str_operation=lambda x: x == 'enabled'),
                    Quantity(
                        'x_psi4_mom',
                        r'MOM (.+)\.', str_operation=lambda x: x == 'enabled'),
                    Quantity(
                        'x_psi4_fractional_occupation',
                        r'Fractional occupation (.+)\.', str_operation=lambda x: x == 'enabled'),
                    Quantity(
                        'x_psi4_guess_type',
                        r'Guess Type is (.+)\.', dtype=str),
                    Quantity(
                        'threshold_energy_change',
                        rf'Energy threshold += +({re_f})', dtype=np.float64, unit=ureg.hartree),
                    Quantity(
                        'threshold_densDFT Potential <==ty_change',
                        rf'Density threDFT Potential <==hold += +({re_f})', dtype=np.float64),
                    Quantity(
                        'x_psi4_integral_threshold',
                        rf'Integral threshold += +({re_f})', dtype=np.float64)])),
            Quantity(
                'basis',
                r'(?:Primary )*Basis (?:Set )*<==([\s\S]+?)==>',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'basis_set',
                        r'((?:Basis Set|Core potential|AO BASIS SET INFORMATION):[\s\S]+?\n\n)',
                        repeats=True, str_operation=str_to_basis)])),
            Quantity(
                'jk_matrices',
                r'(J/K Matrices <==[\s\S]+?==>)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'parameters',
                        r'J/K Matrices <==\s+([\s\S]+?\n *\n)', str_operation=str_to_parameters),
                ])
            ),
            # Move auxiliary out of the parent section because I cannot cover all
            Quantity(
                'auxiliary_basis',
                r' Auxiliary Basis (?:Set )*<=([\s\S]+?)==>',
                repeats=True, sub_parser=TextParser(quantities=[Quantity(
                    'basis_set',
                    r'((?:Basis Set|Core potential|AO BASIS SET INFORMATION):[\s\S]+?\n\n)',
                    repeats=True, str_operation=str_to_basis)])),
            Quantity(
                'dft_potential',
                r'DFT Potential <==([\s\S]+?==>)', sub_parser=TextParser(quantities=[
                    Quantity(
                        'composite',
                        r'Composite Functional.+([\s\S]+?)=>', str_operation=str_to_parameters),
                    Quantity(
                        'exact_exchange',
                        r'Exact \(HF\) Exchange <=([\s\S]+?)=>',
                        str_operation=lambda x: [v.split()[:2] for v in x.strip().split('\n')]),
                    Quantity(
                        'exchange',
                        r'Exchange Functionals <=([\s\S]+?)=>',
                        str_operation=lambda x: [v.split()[:2] for v in x.strip().split('\n')]),
                    Quantity(
                        'correlation',
                        r' Correlation Functionals <=([\s\S]+?)=>',
                        str_operation=lambda x: [v.split()[:2] for v in x.strip().split('\n')]),
                    Quantity(
                        'hybrid',
                        r'Exchange-Correlation Functionals <=([\s\S]+?)=>',
                        str_operation=lambda x: [v.split()[:2] for v in x.strip().split('\n')]),
                    Quantity(
                        'molecular_quadrature',
                        r'Molecular Quadrature <=([\s\S]+?)=>',
                        str_operation=str_to_parameters)])),
            Quantity(
                'method',
                r'(?:\n +(.+) Reference *\n|@(.+) Final Energy:)'),
            Quantity(
                'iterations',
                r'Iterations <==(\s+Total Energy[\s\S]+?)==>',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'iteration',
                        rf'iter +\d+\: +({re_f}) +({re_f}) +({re_f})',
                        repeats=True, dtype=np.float64)])),
            Quantity(
                'spin_contamination_metric',
                rf'@Spin Contamination Metric: +({re_f})', dtype=np.float64),
            Quantity(
                's2_expected',
                rf'@S\^2 Expected: +({re_f})', dtype=np.float64),
            Quantity(
                's2_observed',
                rf'@S\^2 Observed: +({re_f})', dtype=np.float64),
            Quantity(
                's_expected',
                rf'@S   Expected: +({re_f})', dtype=np.float64),
            Quantity(
                's_observed',
                rf'@S   Observed: +({re_f})', dtype=np.float64),
            Quantity(
                'orbital_energies',
                r'(?:Occupied:|Virtual:).+([\s\S]+?\n *\n)',
                repeats=True, str_operation=str_to_orbital_energies),
            Quantity(
                'energy',
                r'Energetics <=([\s\S]+?Total Energy =.+)',
                repeats=True, sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    rf'(.+) Energy = *({re_f})',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)])])),
            Quantity(
                'properties',
                r'(Properties will be evaluated at[\s\S]+?)(?:==>|tstop|\Z)',
                repeats=True, sub_parser=TextParser(quantities=properties_quantities)),
            Quantity(
                'total_gradient',
                r'Total Gradient.+\s+Atom.+\s+\-.+\s+([\s\S]+?)\n *\n',
                str_operation=lambda x: np.array([v.split()[1:] for v in x.strip().split('\n')]),
                dtype=np.dtype(np.float64))]

        mp_quantities = [
            Quantity('method', r'(MP[\d\.]+)\s+Program'),
            Quantity(
                'energy',
                r'Computing MP[\d\.]+ energy using SCF MOs([\s\S]+?)\n *\n',
                repeats=True, sub_parser=TextParser(quantities=[Quantity(
                    'key_value', rf'\s+(.+?) (?:Energy|Contribution) (Correction )*\(a\.u\.\) *: +({re_f})',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)])]))
        ]

        method_quantities = [
            Quantity(
                'parameters',
                r'Parameters <==([\s\S]+?)==>', sub_parser=TextParser(quantities=[Quantity(
                    'key_value', r'([\w ]+= +\S+)',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.split('=')])])),
            Quantity(
                'iterations',
                r'Starting .+? iterations([\s\S]+?)==>',
                sub_parser=TextParser(quantities=[Quantity(
                    'iteration',
                    rf'@.+\d+: +\d+ +({re_f}) +({re_f}) +({re_f})(.+)',
                    repeats=True, dtype=np.dtype(np.float64))])),
            Quantity(
                'energy',
                r'Energetics <==([\s\S]+?Total.+)',
                repeats=True, sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    rf'(.+) [Ee]nergy += *({re_f})',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)])])),
            Quantity(
                'root_info',
                r'(root \d+ information[\s\S]+?==>)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'root_energy',
                        rf'Root \d+ energy += +({re_f})',
                        dtype=np.float64, unit=ureg.hartree)])),
            # TODO parse other properties e.g. orbital extents, mayer bond indices
            Quantity(
                'properties',
                r'(erties computed using the[\s\S]+?)(?:==>|\n *Prop|tstop|\Z)',
                repeats=True, sub_parser=TextParser(quantities=properties_quantities)),
        ]

        mcscf_quantities = [
            Quantity(
                'iterations',
                r'(Cycle +Energy.+\s+\=+[\s\S]+?)\={5}',
                sub_parser=TextParser(quantities=[Quantity(
                    'iteration',
                    rf'@SCF +\d+ +({re_f}) +({re_f}) +({re_f})',
                    repeats=True, dtype=np.dtype(np.float64))])),
            Quantity(
                'energy',
                r'\* (SCF total energy.+)',
                repeats=True, sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    rf'(.+) energy += +({re_f})',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)])])),
            Quantity(
                'orbital_energies',
                r'Eigenvalues \(Eh\)([\s\S]+?)={5}', str_operation=str_to_eigenvalues)
        ]

        cc_quantities = [
            Quantity(
                'parameters',
                r'Wfn Parameters:\s+\-+\s+([\s\S]+?)\n *\n',
                sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    r'(.+ += +\S+)',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.split('=')])])),
            Quantity(
                'energy',
                r'(Nuclear Rep[\s\S]+?)\n *\n',
                repeats=True, sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    rf'([\w\. \-]+) energy += +({re_f})',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)])]))
        ]

        cc_energy_quantities = [
            Quantity(
                'parameters',
                r'Input parameters:\s+\-+([\s\S]+?)',
                sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    r'(.+ += +\S+)',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.split('=')])])),
            Quantity(
                'iterations',
                r'(Iter +Energy[\s\S]+?)\n *\n',
                sub_parser=TextParser(quantities=[Quantity(
                    'iteration',
                    rf'\d+ +({re_f}) +({re_f}) +({re_f})',
                    repeats=True, dtype=np.dtype(np.float64))])),
            Quantity(
                'energy',
                r'\n *\n +(SCF energy[\s\S]+?CCSD total energy.+)',
                repeats=True, sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    rf'([\w\. \-]+) energy.* += +({re_f})',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)])]))
        ]

        cc_lambda_quantities = [
            Quantity(
                'parameters',
                r'Input parameters:\s+\-+([\s\S]+?)\n *\n',
                sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    r'(.+ += +\S+)',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.split('=')])])),
            Quantity(
                'iterations',
                r'(Iter +PseudoEnergy[\s\S]+?)\n *\n',
                sub_parser=TextParser(quantities=[Quantity(
                    'iteration',
                    rf'\d+ +({re_f}) +({re_f})',
                    repeats=True, dtype=np.dtype(np.float64))])),
            Quantity(
                'energy',
                r'(Nuclear Rep\. energy[\s\S]+?Total CCSD energy.+)',
                repeats=True, sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    rf'([\w\. \-]+) energy.* += +({re_f})',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)])]))
        ]

        cc_density_quantities = [
            Quantity(
                'parameters',
                r'Input parameters:\s+\-+([\s\S]+?)\n *\n',
                sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    r'(.+ += +\S+)',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.split('=')])])),
            Quantity(
                'energy',
                r'((?:Nuclear Rep\. energy|Energies re-computed from|Virial Theorem Data)[\s\S]+?)\n *\n',
                repeats=True, sub_parser=TextParser(quantities=[Quantity(
                    'key_value',
                    rf'([\w\. \-]+) energy(?: +\(.+\))* += +({re_f})',
                    repeats=True, str_operation=lambda x: [v.strip() for v in x.rsplit(' ', 1)])])),
            Quantity(
                'properties',
                r'(Properties will be evaluated at[\s\S]+?)(?:==>|tstop|\Z)',
                repeats=True, sub_parser=TextParser(quantities=properties_quantities)),
        ]

        optking_quantities = [
            Quantity(
                'geometry',
                r'Geometry and Gradient-+\s+([\s\S]+?)\n *\n',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'atoms',
                        rf'([A-Z][a-z]*) +({re_f}) +({re_f}) +({re_f})', repeats=True),
                    Quantity(
                        'forces',
                        rf'({re_f}) +({re_f}) +({re_f})',
                        repeats=True, dtype=np.dtype(np.float64))])),
            Quantity(
                'convergence_criteria',
                rf'Convergence Criteria +({re_f}) +\* +({re_f}) +\* +\S+ +({re_f})',
                dtype=np.dtype(np.float64)),
            Quantity('energy', rf'~\s+\d+ +({re_f}) +({re_f})', dtype=np.float64)
        ]

        module_quantities = [
            Quantity(
                'scf',
                rf'(SCF\s+by[\s\S]+?(Computation Completed|tstop|\Z))',
                repeats=True, sub_parser=TextParser(quantities=scf_quantities)),
            Quantity(
                'scf_grad',
                r'(SCF GRAD\s+[\s\S]+?(?:tstop|\Z))',
                repeats=True, sub_parser=TextParser(quantities=scf_quantities)),
            Quantity(
                'mp',
                r'(MP[\d\.]+\s+Program[\s\S]+?(?:tstop|\Z))',
                repeats=True, sub_parser=TextParser(quantities=mp_quantities)),
            Quantity(
                'ci',
                r'(Configuration Interaction\s+\(a [\s\S]+?)(?:PASSED|tstart|\Z)',
                repeats=True, sub_parser=TextParser(quantities=method_quantities)),
            Quantity(
                'mcscf_detci',
                r'(Multi-Configurational Self-Consistent Field[\s\S]+?)(?:PASSED|tstart|\Z)',
                repeats=True, sub_parser=TextParser(quantities=method_quantities)),
            Quantity(
                'mcscf',
                r'(MCSCF: a self-consistent field program[\s\S]+?)(?:tstop|\Z)',
                repeats=True, sub_parser=TextParser(quantities=mcscf_quantities)),
            Quantity(
                'cc',
                r'(Wfn Parameters:\s+\-+[\s\S]+?)(?:tstop|\Z)',
                repeats=True, sub_parser=TextParser(quantities=cc_quantities)),
            Quantity(
                'cc_energy',
                r'(CCENERGY +\*[\s\S]+?)(?:tstop|\Z)',
                repeats=True, sub_parser=TextParser(quantities=cc_energy_quantities)),
            Quantity(
                'cc_lambda',
                r'(CCLAMBDA +\*[\s\S]+?)(?:tstop|\Z)',
                repeats=True, sub_parser=TextParser(quantities=cc_lambda_quantities)),
            Quantity(
                'cc_density',
                r'(CCDENSITY +\*[\s\S]+?)(?:tstop|\Z)',
                repeats=True, sub_parser=TextParser(quantities=cc_density_quantities)),
            Quantity(
                'optking',
                r'OPTKING .+?: for geometry optimizations([\s\S]+?)OPTKING Finished Execution',
                sub_parser=TextParser(quantities=optking_quantities)),
            Quantity(
                'options',
                r'Options:\s+\-+([\s\S]+?)\n *\n',
                str_operation=str_to_parameters),
            Quantity(
                'salc',
                r'(Cartesian Displacement SALCs[\s\S]+?)(?:tstart|\Z)',
                sub_parser=TextParser(quantities=[
                    Quantity(
                        'total_gradient',
                        r'Total gradient:\s+Atom.+\s+.+\s+([\s\S]+?)\n *\n',
                        str_operation=lambda x: [v.split()[1:] for v in x.strip().split('\n')],
                        dtype=np.dtype(np.float64)),
                    Quantity(
                        'contributions_gradient',
                        r'\-(.+ (?:Energy 1st Derivatives|contribution to gradient)):'
                        r'\s+Atom.+\s+.+([\s\S]+?)\n *\n',
                        repeats=True, sub_parser=TextParser(quantities=[
                            Quantity('name', r'(.+) (?:Energy|contribution)', flatten=False),
                            Quantity(
                                'value',
                                rf'\d+ +({re_f}) +({re_f}) +({re_f})',
                                repeats=True, dtype=np.dtype(np.float64))]))])),
        ]

        self._quantities = [
            Quantity(
                'version',
                r'Psi4: An Open-Source Ab Initio Electronic Structure Package\s*'
                r'Psi4 ([\w\.]+)', dtype=str),
            Quantity('start_time', r'Psi4 started on: \w+, (.+)', flatten=False, dtype=str),
            Quantity('process_id', r'Process ID:\s*(\d+)', dtype=np.int32),
            Quantity('psidatadir', r'PSIDATADIR:\s*(.+)', dtype=str),
            Quantity('memory', r'Memory:\s*(\S+) MiB', dtype=np.float64),
            Quantity('threads', r'Threads:\s*(\d+)', dtype=np.int32),
            Quantity('input_file', r'Input File <==\s+\-+\s*([\s\S]+?)\-{5}', flatten=False),
            Quantity(
                'module',
                r'(\*\*\* at [A-Z][a-z]{2} [\s\S]+?(?:\*\*\* tstart|\Z))',
                repeats=True, sub_parser=TextParser(quantities=module_quantities))
        ]


class Psi4Parser:
    def __init__(self):
        self.m_env = m_env
        self.out_parser = OutParser()
        self._xc_map = {
            'HF': 'LDA_X',
            'PW6B95': 'HYB_MGGA_XC_PW6B95',
            'B3LYP': 'HYB_GGA_XC_B3LYP',
            'B3LYP5': 'HYB_GGA_XC_B3LYP5',
            'wB97X-D': 'HYB_GGA_XC_WB97X_D',
        }
        self._metainfo_map = {
            'Nuclear Repulsion': 'nuclear_repulsion',
            'Total': 'total',
            'DFT Exchange-Correlation': 'xc',
            'Nuclear Rep.': 'nuclear_repulsion',
        }

    def init_parser(self):
        self.out_parser.mainfile = self.filepath
        self._method = None

    def parse_system(self, source):
        if source.geometry is None:
            return

        system = self.archive.run[-1].m_create(System)
        atoms = source.geometry.get('atoms', [])
        # TODO determine lattice vectors
        system.atoms = Atoms(
            positions=np.array([a[1:4] for a in atoms]) * ureg.bohr,
            labels=[a[0] for a in atoms], periodic=[False, False, False])
        for key, val in source.geometry.items():
            try:
                setattr(system, 'x_psi4_%s' % key, val)
            except Exception:
                pass

    def parse_method(self, source):
        method = self.archive.run[-1].m_create(Method)
        if self._method is not None:
            method.electronic = Electronic(method=self._method)
            if self._method[:2] in ['MP', 'CI', 'MC', 'CC']:
                method.core_method_ref = self.archive.run[-1].method[0]
                method.methods_ref = [self.archive.run[-1].method[0]]
                if source.parameters is not None:
                    method.x_psi4_parameters = {
                        key: val for key, val in source.parameters.get('key_value', [])}

        def parse_basis(source):
            if source is None:
                return
            basis_set = method.m_create(BasisSet)
            # TODO is this always the case?
            basis_set.type = 'Gaussians'
            for entry in source.get('basis_set', []):
                atom = basis_set.m_create(BasisSetAtomCentered)
                atom.name = entry.get('Basis Set', '').strip('(').strip(')')
                atom.n_basis_functions = int(entry.get('Number of basis function', 0))
                atom.x_psi4_n_shells = int(entry.get('Number of shells', 0))
                atom.x_psi4_max_angular_momentum = int(entry.get('Max angular momentum', 0))
                atom.x_psi4_n_cartesian_functions = int(entry.get('Number of Cartesian functions', 0))
                atom.x_psi4_blend = entry.get('Blend')
                atom.x_psi4_spherical_harmonics = entry.get('Spherical Harmonics?') == 'false'

        # basis set
        parse_basis(source.basis)
        for basis in source.get('auxiliary_basis', []):
            parse_basis(basis)

        if source.jk_matrices is not None:
            method.x_psi4_jk_matrices_parameters = source.jk_matrices.get('parameters', (None, None))[1]

        if source.algorithm is not None:
            scf = method.m_create(ScfMethod)
            for key, val in source.algorithm.items():
                try:
                    setattr(scf, key, val)
                except Exception:
                    pass

        if source.dft_potential is not None:
            dft = method.m_create(DFT)
            xc_functional = dft.m_create(XCFunctional)
            # get xc functional from composite first
            if len(xc_functional.exchange) == 0 and len(xc_functional.correlation) == 0:
                name, parameters = source.dft_potential.get('composite', (None, None))
                if name is not None:
                    name = self._xc_map.get(name)
                    if name is not None:
                        xc_functional.contributions.append(Functional(name=name, parameters=parameters))
            if len(xc_functional.contributions) == 0:
                # get contributions for exchange and correlation
                for xc_type in ['exact_exchange', 'exchange', 'correlation', 'hybrid']:
                    if 'exchange' in xc_type:
                        section = xc_functional.exchange
                    elif xc_type == 'correlation':
                        section = xc_functional.correlation
                    else:
                        section = xc_functional.hybrid
                    for weight, name in source.dft_potential.get(xc_type, []):
                        # it seems that other psi4 version print out xc functionals in libxc
                        # notation prefixed with XC_
                        # TODO complete mapping of XC functionals
                        name = self._xc_map.get(name, name).lstrip('XC_')
                        section.append(Functional(name=name, weight=weight))

            if source.dft_potential.molecular_quadrature is not None:
                dft.x_psi4_molecular_quadrature = source.dft_potential.molecular_quadrature[1]

    def parse_calculation(self, source):
        calc = self.archive.run[-1].m_create(Calculation)
        ref_calcs = []

        def parse_multipole(data, multipoles):
            if len(data) == 3:
                section = multipoles.m_create(MultipolesEntry, Multipoles.dipole)
                exponent = 1
            elif len(data) == 6:
                section = multipoles.m_create(MultipolesEntry, Multipoles.quadrupole)
                exponent = 2
            elif len(data) == 10:
                section = multipoles.m_create(MultipolesEntry, Multipoles.octupole)
                exponent = 3
            elif len(data) == 15:
                section = multipoles.m_create(MultipolesEntry, Multipoles.higher_order)
                section.kind = 'hexadecapole'
                exponent = 4
            else:
                section = multipoles.m_create(MultipolesEntry, Multipoles.higher_order)
                section.kind = '32-pole'
                exponent = 5
            value = data * ureg.elementary_charge * ureg.bohr ** exponent
            section.total = value.to('coulomb * m ** %d' % exponent).magnitude

        def parse_properties(source, calc):
            if source is None:
                return

            # multipoles
            multipole_kinds = ['electronic', 'nuclear', 'total']
            for n, multipole in enumerate(multipole_kinds):
                data = source.get('%s_dipole_moment' % multipole)
                if data is None:
                    continue
                multipoles = calc.m_create(Multipoles) if len(calc.multipoles) <= n else calc.multipoles[n]
                multipoles.kind = multipole_kinds[n]
                parse_multipole(data, multipoles)

            multipole_moments = source.get('multipole_moments', {})
            for multipole in ['dipole', 'quadrupole', 'octupole', 'hexadecapole', 'npole']:
                data = multipole_moments.get(multipole)
                if data is None:
                    continue
                data = np.transpose(data)
                for n, value in enumerate(data):
                    multipoles = calc.m_create(Multipoles) if len(calc.multipoles) <= n else calc.multipoles[n]
                    parse_multipole(value, multipoles)

            # TODO verify for non spin polarized
            # charges
            for method in ['mulliken', 'lowdin']:
                data = source.get('%s_charges' % method)
                if data is not None:
                    charges = calc.m_create(Charges)
                    charges.analysis_method = method
                    if data.atom is not None:
                        charges.value = [a[-1] for a in data.atom] * ureg.elementary_charge
                        charges.spins = [a[-2] for a in data.atom]
                        for n, atom in enumerate(data.atom):
                            for spin in range(2):
                                charges_spin = charges.m_create(ChargesValue, Charges.spin_projected)
                                charges_spin.atom_index = n
                                charges_spin.spin = spin
                                charges_spin.value = atom[spin] * ureg.elementary_charge
                    if data.total is not None:
                        charges.total = data.total[-1]

        # scf iterations
        if source.iterations is not None:
            for iteration in source.iterations.get('iteration', []):
                scf = calc.m_create(ScfIteration)
                scf.energy = Energy(
                    total=EnergyEntry(value=float(iteration[0]) * ureg.hartree),
                    change=float(iteration[1]) * ureg.hartree)

        # energies
        # mp/ci energy, the last entry belongs to the method
        # create additional calc section for references e.g. mp3 have mp2 reference
        for n in range(len(source.get('energy', []))):
            if source.energy[n] is None:
                continue
            ref_calc = calc
            # for MP calc, the last energy block corresponds to calc
            n_ref = len(source.energy) - 1 if self._method[:2] == 'MP' else 0
            if n != n_ref:
                ref_calc = self.archive.run[-1].m_create(Calculation)
                ref_calcs.append(ref_calc)
            ref_calc.energy = Energy()
            for key, val in source.energy[n].get('key_value'):
                key = self._metainfo_map.get(key, key)
                if hasattr(ref_calc.energy, key):
                    setattr(ref_calc.energy, key, EnergyEntry(value=val * ureg.hartree))
                else:
                    ref_calc.energy.contributions.append(EnergyEntry(kind=key, value=val * ureg.hartree))
                # assign total energy to corresponding to method
                if 'total' in key.lower():
                    ref_calc.energy.total = EnergyEntry(value=val * ureg.hartree)

        # forces
        if source.total_gradient is not None:
            calc.forces = Forces(total=ForcesEntry(
                value=source.total_gradient * ureg.hartree / ureg.bohr))

        # spins
        if source.s2_observed is not None:
            calc.x_psi4_s2_expected = source.s2_expected
            calc.x_psi4_s2_observed = source.s2_observed
            calc.spin_S2 = source.s2_observed
            calc.x_psi4_s_expected = source.s_expected
            calc.x_psi4_s_observed = source.s_observed

        # eigenvalues
        if source.orbital_energies is not None:
            eigenvalues = calc.m_create(BandEnergies)
            labels, energies, occupations = [], [], []
            n_spin = len(source.orbital_energies) // 2
            for n, orbital_energies in enumerate(source.orbital_energies):
                labels.extend(orbital_energies[0])
                energies.extend(orbital_energies[1])
                occupations.extend([2 / n_spin if n % 2 == 0 else 0] * len(orbital_energies[0]))
            eigenvalues.orbital_labels = labels
            eigenvalues.energies = np.reshape(energies, (n_spin, 1, len(energies) // n_spin)) * ureg.hartree
            eigenvalues.occupations = np.reshape(occupations, (n_spin, 1, len(occupations) // n_spin))

        # properties can be calculated using multiple density matrices e.g. for ci/mcscf
        # create additional calculation section
        for n, properties in enumerate(source.get('properties', [])):
            ref_calc = calc if n == 0 else self.archive.run[-1].m_create(Calculation)
            if n > 0:
                ref_calcs.append(ref_calc)
            parse_properties(properties, ref_calc)

        # misc
        if source.root_info is not None:
            root_info = calc.m_create(x_psi4_root_information)
            root_info.x_psi4_root_energy = source.root_info.energy

        # reference calculations
        if ref_calcs:
            calc.calculations_ref = ref_calcs
            for ref_calc in ref_calcs:
                ref_calc.starting_calculation_ref = calc

        if calc.m_mod_count == 1:
            self.archive.run[-1].m_remove_sub_section(Run.calculation, -1)

        return calc

    def parse(self, filepath, archive, logger):
        self.filepath = os.path.abspath(filepath)
        self.archive = archive
        self.maindir = os.path.dirname(self.filepath)
        self.logger = logger if logger is not None else logging

        self.init_parser()

        run = archive.m_create(Run)
        run.program = Program(name='Psi4', version=self.out_parser.get('version', ''))
        if self.out_parser.start_time is not None:
            start_time = datetime.strptime(self.out_parser.start_time, '%d %B %Y %H:%M%p')
            run.time_run = TimeRun(date_start=(start_time - datetime(1970, 1, 1)).total_seconds())

        for key in ['process_id', 'psidatadir', 'memory', 'threads', 'input_file']:
            val = self.out_parser.get(key)
            if val is not None:
                setattr(run, 'x_psi4_%s' % key, val)

        opt_calculations_ref = []
        module_names = [
            'scf', 'scf_grad', 'mp', 'ci', 'mcscf', 'mcscf_detci', 'cc', 'cc_energy',
            'cc_lambda', 'cc_density']
        for module in self.out_parser.get('module', []):
            calculations_ref = []
            for name in module_names:
                for submodule in module.get(name, []):
                    self._method = submodule.get('method', name.split('_')[0].upper())
                    self.parse_system(submodule)
                    self.parse_method(submodule)
                    calc = self.parse_calculation(submodule)
                    if calc is not None:
                        if self.archive.run[-1].system:
                            calc.system_ref = self.archive.run[-1].system[-1]
                        if self.archive.run[-1].method:
                            calc.method_ref = self.archive.run[-1].method[-1]
                    calculations_ref.append(calc)
            if module.options is not None:
                self.archive.run[-1].method[-1].x_psi4_options = module.options[1]

            if module.salc is not None:
                calc = self.archive.run[-1].calculation[-1]
                forces = calc.m_create(Forces)
                if module.salc.total_gradient is not None:
                    forces.total = ForcesEntry(
                        value=np.array(module.salc.total_gradient) * ureg.hartree / ureg.bohr)
                for contribution in module.salc.get('contributions_gradient', []):
                    forces.contributions.append(ForcesEntry(
                        kind=contribution.name, value=contribution.value * ureg.hartree / ureg.bohr))

            # each module should can be considered as a single_point workflow
            # TODO implement specific workflow for mp, ci, cc, mcscf
            workflow = self.archive.m_create(Workflow)
            workflow.type = 'single_point'
            workflow.calculations_ref = calculations_ref
            workflow.calculation_result_ref = calc

            if module.optking is not None:
                self.parse_system(module.optking)
                calc = self.archive.run[-1].m_create(Calculation)
                forces = module.optking.get('geometry', {}).get('forces')
                if forces is not None:
                    calc.forces = Forces(total=ForcesEntry(value=forces * ureg.hartree / ureg.bohr))
                energy = module.optking.energy
                if energy is not None:
                    calc.energy = Energy(
                        total=EnergyEntry(value=energy[0] * ureg.hartree),
                        change=energy[1] * ureg.hartree)
                calc.system_ref = self.archive.run[-1].system[-1]
                opt_calculations_ref.append(calc)
        if opt_calculations_ref:
            workflow = self.archive.m_create(Workflow)
            workflow.type = 'geometry_optimization'
            workflow.calculations_ref = opt_calculations_ref
            workflow.calculation_result_ref = opt_calculations_ref[-1]
            convergence = module.optking.convergence_criteria
            if convergence is not None:
                geometry_opt = workflow.m_create(GeometryOptimization)
                geometry_opt.convergence_tolerance_energy_difference = convergence[0] * ureg.hartree
                geometry_opt.convergence_tolerance_force_maximum = convergence[1] * ureg.hartree / ureg.bohr
                geometry_opt.convergence_tolerance_displacement_maximum = convergence[2] * ureg.bohr
